from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Protocol, Set, Tuple, Type

from kolo.types import JSON


class Step(Protocol):
    indent_delta: ClassVar[int] = 0

    def render(self, pytest: bool) -> str: ...  # pragma: no branch

    def get_imports(self, pytest: bool) -> Set["Import"]:
        return set()

    @classmethod
    def get_steps(cls, steps: List["Step"]) -> List[Tuple[slice, List["Step"]]]:
        selected = []
        for index, step in enumerate(steps):
            if isinstance(step, cls):
                selected.append((slice(index, index + 1), [step]))
        return selected


def get_steps_slice(
    start_class: Type[Step], end_class: Type[Step], steps: List[Step]
) -> List[Tuple[slice, List[Step]]]:
    selected = []
    indices = []
    for index, step in enumerate(steps):
        if isinstance(step, start_class):
            indices.append(index)
        elif isinstance(step, end_class):
            start_index = indices.pop()
            _slice = slice(start_index, index + 1)
            selected.append((_slice, steps[_slice]))

    return selected


@dataclass(frozen=True)
class Import:
    import_path: str

    def render(self):
        return f"{self.import_path}\n"


@dataclass()
class CodeComment(Step):
    comment: str

    def render(self, pytest):
        return f"# {self.comment}\n"


@dataclass(frozen=True)
class EmptyLine(Step):
    def render(self, pytest):
        return "\n"


@dataclass(frozen=True)
class ModuleImports(Step):
    def render(self, pytest):
        return ""


@dataclass()
class TestClass(Step):
    name: str
    parents: Tuple[str, ...] = field(default_factory=tuple)
    imports: Set[Import] = field(default_factory=set)
    indent_delta = 1

    def render(self, pytest):
        parents = ", ".join(self.parents)
        return f"class {self.name}({parents}):\n"

    def get_imports(self, pytest):
        return self.imports

    @classmethod
    def get_steps(cls, steps: List[Step]) -> List[Tuple[slice, List[Step]]]:
        return get_steps_slice(cls, EndTestClass, steps)


@dataclass(frozen=True)
class EndTestClass(Step):
    indent_delta = -1

    def render(self, pytest):
        return ""


@dataclass()
class Code(Step):
    code: str
    imports: Set[Import] = field(default_factory=set)

    def render(self, pytest):
        return f"{self.code}\n"

    def get_imports(self, pytest):
        return self.imports


@dataclass()
class TestMethod(Step):
    name: str
    indent_delta = 1

    def render(self, pytest):
        return f"def {self.name}(self):\n"

    @classmethod
    def get_steps(cls, steps: List[Step]) -> List[Tuple[slice, List[Step]]]:
        return get_steps_slice(cls, EndTestMethod, steps)


@dataclass(frozen=True)
class EndTestMethod(Step):
    indent_delta = -1

    def render(self, pytest):
        return ""


@dataclass()
class TestFunction(Step):
    name: str
    fixtures: Tuple[str, ...] = field(default_factory=tuple)
    indent_delta = 1

    def render(self, pytest):
        fixtures = ", ".join(self.fixtures)
        return f"def {self.name}({fixtures}):\n"

    @classmethod
    def get_steps(cls, steps: List[Step]) -> List[Tuple[slice, List[Step]]]:
        return get_steps_slice(cls, EndTestFunction, steps)


@dataclass(frozen=True)
class EndTestFunction(Step):
    indent_delta = -1

    def render(self, pytest):
        return ""


@dataclass()
class StartTimeTravel(Step):
    call: str
    args: Tuple[str, ...] = field(default_factory=tuple)
    imports: Set[Import] = field(default_factory=set)
    indent_delta = 1

    def render(self, pytest):
        args = ", ".join(self.args)
        return f"with {self.call}({args}):\n"

    def get_imports(self, pytest):
        return self.imports

    @classmethod
    def get_steps(cls, steps: List[Step]) -> List[Tuple[slice, List[Step]]]:
        return get_steps_slice(cls, EndTimeTravel, steps)


@dataclass(frozen=True)
class EndTimeTravel(Step):
    indent_delta = -1

    def render(self, pytest):
        return ""


@dataclass()
class With(Step):
    call: str
    args: Tuple[str, ...] = field(default_factory=tuple)
    defines_variable_name: Optional[str] = None
    imports: Set[Import] = field(default_factory=set)
    indent_delta = 1

    def render(self, pytest):
        args = ", ".join(self.args)
        if self.defines_variable_name:
            return f"with {self.call}({args}) as {self.defines_variable_name}:\n"
        return f"with {self.call}({args}):\n"

    def get_imports(self, pytest):
        return self.imports

    @classmethod
    def get_steps(cls, steps: List[Step]) -> List[Tuple[slice, List[Step]]]:
        return get_steps_slice(cls, EndWith, steps)


@dataclass(frozen=True)
class EndWith(Step):
    indent_delta = -1

    def render(self, pytest):
        return ""


@dataclass()
class AssertEqual(Step):
    left: str
    right: str
    imports: Set[Import] = field(default_factory=set)

    def render(self, pytest):
        if pytest:
            return f"assert {self.left} == {self.right}\n"
        return f"self.assertEqual({self.left}, {self.right})\n"

    def get_imports(self, pytest):
        return self.imports


@dataclass()
class AssertStatusCode(Step):
    status_code: int

    def render(self, pytest):
        if pytest:
            return f"assert response.status_code == {self.status_code}\n"
        return f"self.assertEqual(response.status_code, {self.status_code})\n"


@dataclass()
class AssertResponseJson(Step):
    response_json: JSON

    def render(self, pytest):
        if pytest:
            return f"assert response.json() == {self.response_json}\n"
        return f"self.assertEqual(response.json(), {self.response_json})\n"


@dataclass()
class AssertTemplateUsed(Step):
    template_name: str

    def render(self, pytest):
        if pytest:
            assertTemplateUsed = "assertTemplateUsed"
        else:
            assertTemplateUsed = "self.assertTemplateUsed"
        return f"{assertTemplateUsed}(response, {repr(self.template_name)})\n"

    def get_imports(self, pytest):
        if pytest:
            return {Import("from pytest_django.asserts import assertTemplateUsed")}
        return set()


@dataclass()
class RegisterMocket(Step):
    method: str
    url: str
    status_code: int
    body: Optional[str] = None
    json_body: Optional[JSON] = None
    content_type: str = ""

    @classmethod
    def from_outbound_request(cls, outbound_request):
        request = outbound_request["request"]
        response = outbound_request["response"]
        return cls(
            request["method"],
            request["url"],
            response["status_code"],
            response["body"],
            response.get("json_body", None),
            response["content_type"],
        )

    def render(self, pytest):
        rendered = f"""\
Entry.single_register(
    Entry.{self.method},
    "{self.url}",
    status={self.status_code},
"""
        if self.json_body:
            rendered += f"    body=json.dumps({self.json_body}),\n"
        elif self.body:
            rendered += f"    body={repr(self.body)},\n"
        if self.content_type:
            rendered += f'    headers={{"Content-Type": {repr(self.content_type)}}},'
        rendered += ")\n"
        return rendered

    def get_imports(self, pytest):
        imports = {Import("from mocket.mockhttp import Entry")}
        if self.json_body:
            imports.add(Import("import json"))
        return imports


@dataclass()
class DjangoTestClient(Step):
    method: str
    path_info: str
    query_params: Dict[str, Any]
    request_body: Dict[str, Any]
    headers: Dict[str, Any]

    def render(self, pytest):
        if pytest:
            client = "client"
        else:
            client = "self.client"
        rendered = f"""\
response = {client}.{self.method}(
    {repr(self.path_info)},
"""
        if self.request_body:
            rendered += self.render_query_dict(self.request_body)
        elif self.query_params:
            rendered += self.render_query_dict(self.query_params)
        for header, value in self.headers.items():
            rendered += f"    {header}={repr(value)},\n"
        rendered += ")\n"
        return rendered

    def render_query_dict(self, query_dict):
        rendered = "    {"
        for key, value in query_dict.items():
            if isinstance(value, list) and len(value) == 1:
                value = value[0]
            rendered += f"    {repr(key)}: {repr(value)},\n"
        rendered += "    },"
        return rendered


@dataclass(frozen=True)
class DjangoField:
    name: str
    value: str
    imports: Set[Import] = field(default_factory=set)

    @classmethod
    def from_field(cls, field):
        return cls(
            field.name,
            field.value_repr,
            {Import(i) for i in field.imports()},
        )


@dataclass()
class ModelCreate(Step):
    module: str
    model: str
    fields: List[DjangoField]
    defaults: List[DjangoField]
    defines_variable_name: str
    method: str = "get_or_create"

    def get_imports(self, pytest):
        imports = {Import(f"from {self.module} import {self.model}")}
        for field in self.fields:
            imports.update(field.imports)
        for field in self.defaults:
            imports.update(field.imports)
        return imports

    @classmethod
    def from_fixture(cls, fixture):
        return cls(
            module=fixture.module,
            model=fixture.model,
            fields=[DjangoField.from_field(f) for f in fixture.fields],
            defaults=[DjangoField.from_field(fixture.primary_key)],
            defines_variable_name=fixture.variable_name,
        )

    def render(self, pytest):
        rendered = f"{self.defines_variable_name}"
        if self.method == "get_or_create":
            rendered += ", _created"

        rendered += f" = {self.model}.objects.{self.method}(\n"
        for field in self.fields:
            rendered += f"    {field.name}={field.value},\n"

        if self.method == "get_or_create":
            rendered += "    defaults={\n"
            for field in self.defaults:
                rendered += f"        {repr(field.name)}: {field.value},\n"
            rendered += "    },\n"
        rendered += ")\n"
        return rendered


@dataclass()
class ModelUpdate(Step):
    model: str
    fields: List[DjangoField]
    references_variable_name: str

    @classmethod
    def from_fixture(cls, fixture):
        return cls(
            model=fixture.model,
            fields=[DjangoField.from_field(f) for f in fixture.fields],
            references_variable_name=fixture.variable_name,
        )

    def render(self, pytest):
        for field in self.fields:
            rendered = f"{self.references_variable_name}.{field.name} = {field.value}\n"

        rendered += f"{self.references_variable_name}.save()\n"
        return rendered


@dataclass()
class FactoryCreate(Step):
    module: str
    factory: str
    fields: List[DjangoField]
    defines_variable_name: str

    @classmethod
    def from_fixture(cls, fixture):
        return cls(
            module=fixture.module,
            factory=fixture.factory,
            fields=[DjangoField.from_field(f) for f in fixture.fields],
            defines_variable_name=fixture.variable_name,
        )

    def render(self, pytest):
        rendered = f"{self.defines_variable_name} = {self.factory}.create(\n"
        for field in self.fields:
            rendered += f"    {field.name}={field.value},\n"
        rendered += ")\n"
        return rendered

    def get_imports(self, pytest):
        imports = {Import(f"from {self.module} import {self.factory}")}
        for field in self.fields:
            imports.update(field.imports)
        return imports


@dataclass()
class AssertInsert(Step):
    module: str
    model: str
    lookup_fields: List[DjangoField]
    assert_fields: List[DjangoField]
    defines_variable_name: str

    @classmethod
    def from_fixture(cls, fixture):
        lookup_fields, assert_fields = fixture.get_fields()
        return cls(
            module=fixture.module,
            model=fixture.model,
            lookup_fields=[DjangoField.from_field(f) for f in lookup_fields],
            assert_fields=[DjangoField.from_field(f) for f in assert_fields],
            defines_variable_name=fixture.variable_name,
        )

    def render(self, pytest):
        rendered = f"{self.defines_variable_name} = {self.model}.objects.get(\n"
        for field in self.lookup_fields:
            rendered += f"    {field.name}={field.value},\n"
        rendered += ")\n"
        for field in self.assert_fields:
            if pytest:
                rendered += f"assert {self.defines_variable_name}.{field.name} == {field.value}\n"
            else:
                rendered += f"self.assertEqual({self.defines_variable_name}.{field.name}, {field.value})\n"

        return rendered

    def get_imports(self, pytest):
        imports = {Import(f"from {self.module} import {self.model}")}
        for field in self.lookup_fields:
            imports.update(field.imports)
        for field in self.assert_fields:
            imports.update(field.imports)
        return imports


@dataclass()
class AssertUpdate(Step):
    model: str
    fields: List[DjangoField]
    references_variable_name: str

    @classmethod
    def from_fixture(cls, fixture):
        return cls(
            model=fixture.model,
            fields=[DjangoField.from_field(f) for f in fixture.fields],
            references_variable_name=fixture.variable_name,
        )

    def render(self, pytest):
        rendered = f"{self.references_variable_name}.refresh_from_db()\n"
        for field in self.fields:
            if pytest:
                rendered += f"assert {self.references_variable_name}.{field.name} == {field.value}\n"
            else:
                rendered += f"self.assertEqual({self.references_variable_name}.{field.name}, {field.value})\n"
        return rendered

    def get_imports(self, pytest):
        imports = set()
        for field in self.fields:
            imports.update(field.imports)
        return imports


@dataclass()
class AssertDelete(Step):
    module: str
    model: str
    fields: List[DjangoField]

    @classmethod
    def from_fixture(cls, fixture):
        return cls(
            module=fixture.module,
            model=fixture.model,
            fields=[DjangoField.from_field(f) for f in fixture.fields],
        )

    def grouped_fields(self):
        fields: Dict[str, List[str]] = {}
        for field in self.fields:
            fields.setdefault(field.name, []).append(field.value)
        return fields

    def render(self, pytest):
        if pytest:
            rendered = "assert not "
        else:
            rendered = "self.assertFalse("
        rendered += f"{self.model}.objects.filter(\n"

        for name, values in self.grouped_fields().items():
            if len(values) == 1:
                value = values[0]
                rendered += f"    {name}={value},\n"
            else:
                joined_values = ", ".join(values)
                rendered += f"    {name}__in=({joined_values}),\n"
        if pytest:
            rendered += ").exists()\n"
        else:
            rendered += ").exists())\n"
        return rendered

    def get_imports(self, pytest):
        imports = {Import(f"from {self.module} import {self.model}")}
        for field in self.fields:
            imports.update(field.imports)
        return imports



from __future__ import annotations

import json
import os

from ..config import create_kolo_directory
from ..settings import get_webapp_settings
from ..version import __version__


def web_home_html(running_in_user_django: bool):
    kolo_dir = create_kolo_directory()
    project_folder = kolo_dir.parent
    settings = get_webapp_settings()

    source_dir = os.path.dirname(__file__)
    with open(f"{source_dir}/templates/home.html") as f:
        html = f.read().format(
            settings=json.dumps(settings),
            __version__=json.dumps(__version__),
            project_folder=json.dumps(str(project_folder)),
            running_in_user_django=json.dumps(running_in_user_django),
        )
    return html



from hashlib import sha1

from django.http import HttpRequest, HttpResponse, JsonResponse
from django.utils.cache import patch_response_headers
from django.utils.dateparse import parse_datetime
from django.utils.http import http_date
from django.views.decorators.cache import cache_control
from django.views.decorators.http import condition, conditional_page

from ..db import get_db_last_modified
from .api import delete_trace, generate_test, get_trace, latest_traces, read_source_file
from .home import web_home_html


def kolo_web_home(request: HttpRequest) -> HttpResponse:
    return HttpResponse(web_home_html(running_in_user_django=True))


def get_last_modified(request):
    """
    Return the last modified time of the db.

    Cache the value on request to avoid duplicate work in
    kolo_web_api_latest_traces_etag and kolo_web_api_latest_traces_last_modified
    which can be called in either order based on the Django version.

    The order changed in Django 5.0:
    https://github.com/django/django/commit/d3d173425fc0a1107836da5b4567f1c88253191b
    """
    try:
        return request._db_last_modified
    except AttributeError:
        request._db_last_modified = get_db_last_modified()
    return request._db_last_modified


def kolo_web_api_latest_traces_etag(request, *a, **kwargs):
    last_modified = get_last_modified(request)
    if last_modified is not None:
        return sha1(last_modified.isoformat().encode("utf-8")).hexdigest()
    return None


def kolo_web_api_latest_traces_last_modified(request, *a, **kwargs):
    return get_last_modified(request)


@cache_control(max_age=1)
@condition(
    etag_func=kolo_web_api_latest_traces_etag,
    last_modified_func=kolo_web_api_latest_traces_last_modified,
)
def kolo_web_api_latest_traces(request: HttpRequest) -> HttpResponse:
    if "anchor" in request.GET and "showNext" in request.GET:
        anchor = request.GET["anchor"]
        show_next = int(request.GET["showNext"])
    else:
        anchor = None
        show_next = int(request.GET.get("showNext", 10))

    return JsonResponse(latest_traces(show_next, anchor))


@conditional_page
def kolo_web_api_get_trace(request: HttpRequest) -> HttpResponse:
    trace_id = request.path.replace("/_kolo/api/traces/", "").replace("/", "")
    msgpack_data, created_at = get_trace(trace_id)
    response = HttpResponse(msgpack_data, content_type="application/msgpack")
    # When Chrome (Blink) and Safari (WebKit) do a `fetch()` request, they include the `If-None-Match` and
    # `If-Modified-Since` headers which allow for `ETag` (or `Last-Modified`) matching and returning a
    # 304 (Not Modified) response. FireFox (Gecko) only does so if a {cache: "force-cache"} value is given as options.
    last_modified = parse_datetime(created_at)
    if last_modified is not None:
        response["Last-Modified"] = http_date(last_modified.timestamp())
    response["ETag"] = trace_id
    # Once we've seen the trace, don't even try asking for the URL again for a while (a month) - because the traces are
    # ostensibly immutable, we should only really expect to *need* to ask for them again upon cache eviction
    patch_response_headers(response, cache_timeout=86400 * 28)
    return response


def kolo_web_api_delete_trace(request: HttpRequest) -> HttpResponse:
    trace_id = request.path.replace("/_kolo/api/traces/", "").replace("/", "")
    return JsonResponse(delete_trace(trace_id))


def kolo_web_api_generate_test(request: HttpRequest) -> HttpResponse:
    trace_id = request.path.replace("/_kolo/api/generate-test/", "").replace("/", "")
    return JsonResponse(generate_test(trace_id))


def kolo_web_api_source_file(request: HttpRequest) -> HttpResponse:
    file_path = request.GET["path"]
    return JsonResponse(read_source_file(file_path))



from datetime import datetime
from typing import Any

import sqlglot
from more_itertools import flatten


def sql_overview(query, *, database, query_data=None):
    parsed = sqlglot.parse_one(query, read=database)
    data: Any = {}
    if isinstance(parsed, sqlglot.expressions.Select):
        if query_data is None:
            return {"type": "other"}

        data["type"] = "select"
        data["columns"] = []

        try:
            rows = list(flatten(query_data))
        except TypeError:
            return {"type": "other"}

        for row in rows:
            columns = []
            for column, value in zip(parsed.expressions, row):
                if isinstance(column, sqlglot.expressions.Column):
                    column_data = {
                        "name": column.name,
                        "table": column.table,
                        "derived": False,
                        "value": value,
                    }
                else:
                    children = column.find_all(sqlglot.expressions.Column)
                    tables = {column.table for column in children}
                    column_data = {
                        "name": column.alias_or_name,
                        "tables": tables,
                        "derived": True,
                        "value": value,
                    }
                columns.append(column_data)
            data["columns"].append(columns)

    elif isinstance(parsed, sqlglot.expressions.Insert):
        data["type"] = "insert"
        data["table"] = parsed.this.this.name
        data["columns"] = []
        for values in parsed.expression.expressions:
            columns = [
                {
                    "name": column.name,
                    "value": parse_value(value),
                    "table": data["table"],
                }
                for column, value in zip(parsed.this.expressions, values)
            ]
            data["columns"].append(columns)
        data["returning"] = []
        if parsed.args["returning"]:
            for row in query_data:
                returning = []
                for column, value in zip(parsed.args["returning"].expressions, row):
                    returning.append(
                        {
                            "name": column.name,
                            "value": value,
                        }
                    )
                data["returning"].append(returning)

    elif isinstance(parsed, sqlglot.expressions.Update):
        data["type"] = "update"
        data["table"] = parsed.this.this.name
        data["where"] = parse_where(parsed)
        data["columns"] = [
            {
                "name": expression.this.name,
                "value": parse_value(expression.expression),
                "table": data["table"],
            }
            for expression in parsed.expressions
        ]

    elif isinstance(parsed, sqlglot.expressions.Delete):
        data["type"] = "delete"
        tables = parsed.args["tables"]
        if tables:
            data["tables"] = {table.name for table in tables}
        else:
            data["tables"] = {parsed.this.name}

        data["where"] = parse_where(parsed)

    else:
        data["type"] = "other"

    return data


def parse_where(parsed):
    where = parsed.args["where"]
    if where is None:
        return {}

    where = where.this
    if isinstance(where, sqlglot.expressions.EQ):
        return {
            "name": where.this.name,
            "value": where.expression.name,
        }
    elif isinstance(where, sqlglot.expressions.In):
        return {
            "name": where.this.name,
            "value": [v.name for v in where.expressions],
        }
    raise ValueError(f"Unexpected WHERE expression: {where}")  # pragma: no cover


def parse_value(expression):
    if isinstance(expression, sqlglot.expressions.Literal):
        return expression.name
    if isinstance(expression, sqlglot.expressions.Boolean):
        return expression.this
    if isinstance(expression, sqlglot.expressions.Cast):
        value = parse_value(expression.this)
        if expression.to.this == sqlglot.expressions.DataType.Type.TIMESTAMPTZ:
            return datetime.fromisoformat(value)
    if isinstance(expression, sqlglot.expressions.Null):
        return None
    return str(expression)



import json
from typing import Literal, TypedDict

from . import kv

WEBAPP_SETTINGS_KEY = "webapp_settings"


class WebappSettings(TypedDict):
    auto_delete_old_traces: bool
    default_text_editor: Literal["vscode", "pycharm"]


default_webapp_settings: WebappSettings = {
    "auto_delete_old_traces": True,
    "default_text_editor": "vscode",
}


def get_webapp_settings() -> WebappSettings:
    try:
        value = kv.get_value(WEBAPP_SETTINGS_KEY)
    except KeyError:
        return default_webapp_settings
    saved_settings: WebappSettings = json.loads(value) if value else {}  # type: ignore
    return {**default_webapp_settings, **saved_settings}


def save_webapp_settings(settings: WebappSettings):
    kv.set(WEBAPP_SETTINGS_KEY, json.dumps(settings))



from __future__ import annotations

import types

from ..serialize import decode_body, decode_header_value


def get_full_url(frame_locals) -> str:
    scheme = frame_locals["self"].scheme
    host = frame_locals["self"].host
    port = frame_locals["self"].port
    if (
        port is None
        or (scheme == "https" and port == 443)
        or (scheme == "http" and port == 80)
    ):
        port = ""
    else:
        port = f":{port}"
    url = frame_locals["url"]
    return f"{scheme}://{host}{port}{url}"


def process_urllib3(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    frame_locals = frame.f_locals
    full_url = get_full_url(frame_locals)
    method = frame_locals["method"].upper()
    method_and_full_url = f"{method} {full_url}"

    if event == "call":
        request_headers = {
            key: decode_header_value(value)
            for key, value in frame_locals["headers"].items()
        }

        return {
            "body": decode_body(frame_locals["body"], request_headers),
            "headers": request_headers,
            "method": method,
            "method_and_full_url": method_and_full_url,
            "url": full_url,
        }

    assert event == "return"

    try:
        response = frame_locals["response"]
    except KeyError:
        headers = None
        status = None
    else:
        headers = dict(response.headers)
        status = response.status

    return {
        "body": None,
        "headers": headers,
        "method": method,
        "method_and_full_url": method_and_full_url,
        "status_code": status,
        "url": full_url,
    }


urllib3 = {
    "co_names": ("urlopen",),
    "path_fragment": "urllib3/connectionpool",
    "call_type": "outbound_http_request",
    "return_type": "outbound_http_response",
    "subtype": "urllib3",
    "process": process_urllib3,
}



import json
import re
import uuid
from ast import literal_eval
from collections import Counter
from datetime import date, datetime, time, timedelta, timezone
from decimal import Decimal
from enum import Enum, auto
from functools import total_ordering
from itertools import chain, product
from textwrap import dedent
from typing import Any, Dict, List, Set, Tuple, Union

import sqlglot
from more_itertools import unique_everseen

from kolo.django_schema import verbose_name

from .loaders import import_string
from .sort import TopologicalSorter

DATABASES = {
    "postgresql": "postgres",
}


UUID_REPR_REGEX = re.compile(
    r"UUID\(\'([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})\'\)"
)
UUID_STR_REGEX = re.compile(
    r"([0-9a-f]{8}-?[0-9a-f]{4}-?[0-9a-f]{4}-?[0-9a-f]{4}-?[0-9a-f]{12})"
)
DECIMAL_REPR_REGEX = re.compile(r"Decimal\('(\d+.?\d+?)'\)")
TIMEDELTA_REPR_REGEX = re.compile(
    r"datetime.timedelta\((?:days=(\d+),? ?)?(?:seconds=(\d+),? ?)?(?:microseconds=(\d+))?\)"
)
TIMEDELTA_STR_REGEX = re.compile(r"(-?\d+) days (\d+)\.(\d+) seconds")
TIMEDELTA_MICROSECONDS_REGEX = re.compile(r"\d+")


def is_auto_now(field):
    django_field = field.schema["django_field"]
    if django_field not in (
        "django.db.models.fields.DateTimeField",
        "django.db.models.fields.DateField",
    ):
        return False
    return field.schema["auto_now"] or field.schema["auto_now_add"]


class TooComplicatedError(Exception):
    pass


class SkipQueryReason(Enum):
    NO_COLUMNS = auto()
    NO_PRIMARY_KEY = auto()
    NULL_PRIMARY_KEY = auto()
    SEEN_MUTATION = auto()
    UNSUPPORTED_WHERE = auto()


class SkipQuery:
    def __init__(self, query: Dict[str, Any], reason: SkipQueryReason, table=None):
        self.query = query
        self.reason = reason
        self.table = table

    def __repr__(self):
        return f"{self.reason} table: {self.table}"


class AssertInsert:
    def __init__(self, variable_name, module, model, table, fields, query=None):
        self.variable_name = variable_name
        self.module = module
        self.model = model
        self.table = table
        self.fields = fields
        self.query = query

    @classmethod
    def from_raw(cls, fields, query, table_schema, table_name, names, primary_key):
        module = table_schema["model_module"]
        model = table_schema["model_name"]
        verbose_name = table_schema["verbose_name"]
        variable_name = names.next_variable_name(
            verbose_name, primary_key.name, primary_key.value_repr
        )
        return cls(
            variable_name,
            f"{module}",
            f"{model}",
            table_name,
            fields,
            query,
        )

    def update_fields(self, names, schema):
        fields = []
        for field in self.fields:
            if field.schema["primary_key"]:
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.DateTimeField":
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.DurationField":
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.json.JSONField":
                continue  # pragma: no cover
            if field.skip_default():
                continue

            if field.schema["is_relation"]:
                related_model = field.schema["related_model"]
                verbose_name = schema[related_model]["verbose_name"]
                related_pk = field.schema["related_pk"]
                try:
                    field.value_repr = field.value = names.names[
                        (verbose_name, related_pk, field.value_repr)
                    ]
                except KeyError:  # pragma: no cover
                    pass

            fields.append(field)

        for name, schema in schema[self.table]["fields"].items():
            if schema["gfk_object_id_field"]:
                fields = collect_gfk(fields, name, schema, names)
        self.fields = fields

    def __eq__(self, other):
        if not isinstance(other, AssertInsert):
            return NotImplemented  # pragma: no cover
        return (self.variable_name, self.module, self.model, self.fields) == (
            other.variable_name,
            other.module,
            other.model,
            other.fields,
        )

    def __repr__(self):
        return dedent(  # pragma: no cover
            f"""\
            AssertInsert(
                "{self.variable_name}",
                "{self.module}",
                "{self.model}",
                "{self.table}",
                {self.fields},
            )"""
        )

    def unique_fields(self):
        unique = []
        rest = []
        for field in self.fields:
            if field.schema.get("has_unique", False):
                unique.append(field)
            else:
                rest.append(field)
        return unique, rest

    def foreign_keys(self):
        foreign = []
        rest = []
        for field in self.fields:
            if field.schema["is_relation"]:
                foreign.append(field)
            else:
                rest.append(field)
        return foreign, rest

    def get_fields(self):
        fields, rest = self.unique_fields()
        if fields:
            return fields, rest

        fields, rest = self.foreign_keys()
        if fields:
            return fields, rest

        return self.fields, []


class AssertUpdate:
    def __init__(self, variable_name, model, table, fields, query=None):
        self.variable_name = variable_name
        self.model = model
        self.table = table
        self.fields = fields
        self.query = query

    def __repr__(self):
        return dedent(
            f"""\
            AssertUpdate(
                "{self.variable_name}",
                "{self.model}",
                "{self.table}",
                {self.fields},
            )"""
        )

    def __eq__(self, other):
        if not isinstance(other, AssertUpdate):
            return NotImplemented  # pragma: no cover
        return (self.variable_name, self.fields) == (other.variable_name, other.fields)

    def update_fields(self, names, schema):
        fields = []
        for field in self.fields:
            if field.schema["primary_key"]:
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.DateTimeField":
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.json.JSONField":
                continue  # pragma: no cover
            if field.skip_default():
                continue

            if field.schema["is_relation"]:
                related_model = field.schema["related_model"]
                verbose_name = schema[related_model]["verbose_name"]
                related_pk = field.schema["related_pk"]
                try:
                    field.value_repr = field.value = names.names[
                        (verbose_name, related_pk, field.value_repr)
                    ]
                except KeyError:  # pragma: no cover
                    pass
                else:
                    field.name = field.schema["field_name"]

            fields.append(field)

        for name, schema in schema[self.table]["fields"].items():
            if schema["gfk_object_id_field"]:
                fields = collect_gfk(fields, name, schema, names)
        self.fields = fields


class AssertDelete:
    def __init__(self, module, model, fields, query=None):
        self.module = module
        self.model = model
        self.fields = fields
        self.query = query

    def __repr__(self):
        return dedent(
            f"""\
            AssertDelete(
                "{self.module}",
                "{self.model}",
                {self.fields},
            )"""
        )

    def __eq__(self, other):
        if not isinstance(other, AssertDelete):
            return NotImplemented  # pragma: no cover
        return (self.module, self.model, self.fields) == (
            other.module,
            other.model,
            other.fields,
        )

    def update_fields(self, names, schema):
        pass

    def grouped_fields(self):
        fields: Dict[str, List[DjangoField]] = {}
        for field in self.fields:
            fields.setdefault(field.name, []).append(field)
        return fields


class DjangoUpdate:
    def __init__(
        self,
        table,
        model,
        fields,
        variable_name,
    ):
        self.table = table
        self.model = model
        self.fields = fields
        self.variable_name = variable_name

    def __eq__(self, other):
        if not isinstance(other, DjangoUpdate):
            return NotImplemented
        return (self.table, self.model, self.fields, self.variable_name) == (
            other.table,
            other.model,
            other.fields,
            other.variable_name,
        )

    def __repr__(self):
        return dedent(
            f"""\
            DjangoUpdate(
                "{self.table}",
                "{self.model}",
                {self.fields},
                "{self.variable_name}",
            )"""
        )

    def update_fields(self, names, schema):
        fields = self.fields

        updated_fields = []
        for field in fields:
            if field.schema["null"] and field.value is None:
                continue

            related_model = field.schema["related_model"]
            verbose_name = schema[related_model]["verbose_name"]
            related_pk = field.schema["related_pk"]
            field.value_repr = field.value = names.names[
                (verbose_name, related_pk, field.value_repr)
            ]
            updated_fields.append(field)

        self.fields = tuple(updated_fields)

    def is_empty(self):
        return not bool(self.fields)


class DjangoCreate:
    def __init__(
        self,
        table,
        module,
        model,
        primary_key,
        fields,
        variable_name,
        query=None,
    ):
        self.table = table
        self.module = module
        self.model = model
        self.primary_key = primary_key
        self.fields = fields
        self.query = query
        self.variable_name = variable_name

    @classmethod
    def from_raw(cls, table, fields, query, schema_data, names):
        module = schema_data[table]["model_module"]
        model = schema_data[table]["model_name"]
        fields = tuple(fields)
        primary_key = find_pk(fields)
        verbose_name = schema_data[table]["verbose_name"]
        variable_name = names.next_variable_name(
            verbose_name, primary_key.name, primary_key.value_repr
        )
        fields, update_fields = find_update_fields(fields, table)
        update = None
        if update_fields:
            update = DjangoUpdate(table, f"{model}", update_fields, variable_name)

        return (
            cls(
                table,
                f"{module}",
                f"{model}",
                primary_key,
                fields,
                variable_name,
                query,
            ),
            update,
        )

    def all_fields(self):
        yield self.primary_key
        yield from self.fields

    def update_fields(self, names, schema):
        fields = self.fields
        for name, field_schema in schema[self.table]["fields"].items():
            if field_schema["gfk_object_id_field"]:
                fields = collect_gfk(fields, name, field_schema, names)

        updated_fields = []
        for field in fields:
            if field.schema["primary_key"]:
                continue
            if field.schema["null"] and field.value is None:
                continue
            if is_auto_now(field):
                continue
            if field.skip_default():
                continue
            if field.schema["is_relation"]:
                related_model = field.schema["related_model"]
                if related_model is None:
                    updated_fields.append(field)
                    continue  # Generic Foreign Key
                verbose_name = schema[related_model]["verbose_name"]
                related_pk = field.schema["related_pk"]
                try:
                    field.value_repr = field.value = names.names[
                        (verbose_name, related_pk, field.value_repr)
                    ]
                except KeyError:
                    field.name = field.schema["field_attname"]
            updated_fields.append(field)

        self.fields = tuple(updated_fields)

    def is_empty(self):
        return False

    def __eq__(self, other):
        if not isinstance(other, DjangoCreate):
            return NotImplemented
        return (self.table, self.model, self.fields, self.variable_name) == (
            other.table,
            other.model,
            other.fields,
            other.variable_name,
        )

    def __repr__(self):
        return dedent(
            f"""\
            DjangoCreate(
                "{self.table}",
                "{self.module}",
                "{self.model}",
                {self.primary_key},
                {self.fields},
                "{self.variable_name}",
            )"""
        )


def extract_value(expression):
    if isinstance(expression, sqlglot.expressions.Literal):
        if expression.is_string:
            return expression.name
        return literal_eval(expression.name)
    if isinstance(expression, sqlglot.expressions.Boolean):
        return expression.this
    if isinstance(expression, sqlglot.expressions.Null):
        return None  # pragma: no cover
    if isinstance(expression, sqlglot.expressions.Cast) and expression.name:
        return expression.name
    raise TooComplicatedError()


@total_ordering
class DjangoField:
    def __init__(self, name, column_name, value, schema=None):
        self.name = name
        self.column_name = column_name
        self.value_repr = repr(value)
        self.raw_value = value
        self.value = repr(value) if isinstance(value, str) else value
        self.schema = schema

    @classmethod
    def from_column(cls, column, value, table_schema, parser):
        field_schema = table_schema["fields"][column.name]
        name = field_schema["field_name"]
        value = parser.parse_value(value, field_schema)
        return cls(name, column.name, value, field_schema)

    @classmethod
    def from_identifier(cls, identifier, value, table_schema, parser):
        field_schema = table_schema["fields"][identifier.name]
        if isinstance(value, sqlglot.expressions.Null):
            value = None
        elif isinstance(value, sqlglot.expressions.Boolean):
            value = value.this
        elif isinstance(value, sqlglot.expressions.Cast):
            value = value.name
        elif isinstance(value, sqlglot.expressions.Anonymous):
            value = str(value)
        elif value.is_string:
            value = value.name
        else:
            django_field = field_schema["django_field"]
            if django_field == "django.db.models.fields.DecimalField":
                value = value.name
            else:
                value = literal_eval(value.name)

            if django_field == "django.db.models.fields.BooleanField" and isinstance(
                value, int
            ):
                value = bool(value)
        value = parser.parse_value(value, field_schema)
        return cls(field_schema["field_name"], identifier.name, value, field_schema)

    @classmethod
    def from_eq(cls, eq, schema_fields, parser, lookup_name, lookup_value):
        name = eq.left.name
        right = eq.right
        if isinstance(right, sqlglot.expressions.Cast):
            right = right.this
            while isinstance(right, sqlglot.expressions.Paren):
                right = right.this

        if isinstance(right, sqlglot.expressions.Case):
            whens = [
                e
                for e in right.iter_expressions()
                if isinstance(e, sqlglot.expressions.If)
            ]
            for when in whens:  # pragma: no branch
                when_eq = when.find(sqlglot.expressions.EQ)
                if when_eq is None:
                    raise TooComplicatedError()  # pragma: no cover
                column = when_eq.this
                expression = when_eq.expression
                if column.name == lookup_name and expression.name == lookup_value:
                    value = extract_value(when.args["true"])
                    break
            else:  # pragma: no cover
                defaults = [
                    e
                    for e in right.iter_expressions()
                    if not isinstance(e, sqlglot.expressions.If)
                ]
                value = extract_value(defaults[0])
        else:
            value = extract_value(right)

        field_schema = schema_fields[name]
        value = parser.parse_value(value, field_schema)
        return cls(field_schema["field_name"], name, value, field_schema)

    def __eq__(self, other):
        return (self.name, self.column_name, self.value) == (
            other.name,
            other.column_name,
            other.value,
        )

    def __lt__(self, other) -> bool:
        if self.name == other.name:
            if self.value is None:
                return False
            if other.value is None:
                return True
            return repr(self.value) < repr(other.value)
        return self.name < other.name

    def __repr__(self):
        return f'DjangoField("{self.name}", "{self.column_name}", {self.value_repr})'

    def skip_default(self):
        if "default" not in self.schema:
            return False

        default_schema = self.schema["default"]
        if repr(default_schema["value"]) == self.value_repr:
            return True

        if "type" not in default_schema:
            return False

        field_type = default_schema["type"]
        value = default_schema["value"]
        if field_type == "Decimal" and self.value == Decimal(value):
            return True
        if field_type == "timedelta" and self.value == timedelta(  # pragma: no branch
            **value
        ):
            return True
        return False

    def imports(self) -> Set[str]:
        if isinstance(self.value, Decimal):
            return {"from decimal import Decimal"}
        elif isinstance(self.value, (date, datetime, time, timedelta, timezone)):
            return {"import datetime"}
        elif isinstance(self.value, uuid.UUID):
            return {"from uuid import UUID"}
        return set()


class VariableNames:
    def __init__(self):
        self.counts: Counter = Counter()
        self.names = {}

    def next_variable_name(self, verbose_name, field_name, field_repr):
        full_key = verbose_name, field_name, field_repr
        try:
            return self.names[full_key]
        except KeyError:
            pass

        key = verbose_name, field_name
        self.counts[key] += 1
        suffix = self.counts[key]
        if suffix == 1:
            name = verbose_name
        else:
            name = f"{verbose_name}_{suffix}"
        self.names[full_key] = name
        return name


class SQLParser:
    def __init__(self, config):
        parsers = config.get("test_generation", {}).get("field_parsers", [])
        self.field_parsers = tuple(map(import_string, parsers))
        self.names = VariableNames()

    def parse_value(self, value, column_schema):
        if value is None:
            return None
        django_field = column_schema["django_field"]
        if django_field == "django.db.models.fields.UUIDField":
            uuid_repr_match = UUID_REPR_REGEX.match(str(value))
            uuid_str_match = UUID_STR_REGEX.match(str(value))
            if uuid_repr_match:
                value = uuid_repr_match[1]
                value = uuid.UUID(value)
            elif uuid_str_match:
                value = uuid_str_match[1]
                value = uuid.UUID(value)
        elif django_field == "django.db.models.fields.json.JSONField":
            try:
                value = json.loads(value)
            except json.decoder.JSONDecodeError:
                pass
        elif django_field == "django.db.models.fields.DecimalField":
            decimal_match = DECIMAL_REPR_REGEX.match(str(value))
            if decimal_match:
                value = Decimal(decimal_match[1])
            elif isinstance(value, str):
                value = Decimal(value)
            if isinstance(value, Decimal):
                value = value.normalize()
        elif django_field == "django.db.models.fields.DurationField":
            if isinstance(value, str):
                timedelta_match = TIMEDELTA_REPR_REGEX.match(value)
                duration_match = TIMEDELTA_STR_REGEX.match(value)
                microseconds_match = TIMEDELTA_MICROSECONDS_REGEX.match(value)
                if timedelta_match:
                    groups = [int(g) if g else 0 for g in timedelta_match.groups()]
                    value = timedelta(
                        days=groups[0], seconds=groups[1], microseconds=groups[2]
                    )
                elif duration_match:
                    groups = duration_match.groups()  # type: ignore
                    value = timedelta(
                        days=int(groups[0]),
                        seconds=int(groups[1]),
                        microseconds=int(groups[2]),
                    )
                elif microseconds_match:  # pragma: no branch
                    value = timedelta(microseconds=int(microseconds_match.group()))
            else:
                value = timedelta(microseconds=value)
        elif django_field == "django.db.models.fields.DateTimeField":
            if isinstance(value, str):
                value = datetime.fromisoformat(value)
            value = datetime.combine(
                value.date(), value.time(), value.tzinfo or timezone.utc
            )
        elif django_field == "django.db.models.fields.DateField":
            if isinstance(value, str):
                value = date.fromisoformat(value)
        else:
            for parser in self.field_parsers:
                value = parser(value, django_field)
        return value

    def build_assert_inserts(
        self, parsed_query, query, table_schema, table_name, seen_mutations
    ):
        sql_schema = parsed_query.find(sqlglot.expressions.Schema)
        if sql_schema is None:
            return  # pragma: no cover
        schema_columns = [
            column
            for column in sql_schema.iter_expressions()
            if isinstance(column, sqlglot.expressions.Identifier)
        ]
        returning = parsed_query.find(sqlglot.expressions.Returning)
        values = parsed_query.find(sqlglot.expressions.Values)
        for sql_values_tuple in values.expressions:
            value_columns = [column for column in sql_values_tuple.iter_expressions()]
            zipped_columns = list(zip(schema_columns, value_columns))
            fields = self.build_insert_assert_fields(zipped_columns, table_schema)
            primary_key = None
            if returning:  # pragma: no branch
                for field, value in zip(returning.expressions, query["query_data"][0]):
                    field_schema = table_schema["fields"][field.name]
                    if field_schema["primary_key"]:  # pragma: no branch
                        primary_key = DjangoField(
                            field.this.name,
                            field.this.name,
                            value,
                            field_schema,
                        )
                        seen_mutations.setdefault(table_name, set()).add(repr(value))
            if primary_key is None:  # pragma: no cover
                continue
            yield AssertInsert.from_raw(
                fields, query, table_schema, table_name, self.names, primary_key
            )

    def build_insert_assert_fields(self, columns, table_schema):
        fields = []
        for identifier, value in columns:
            field = DjangoField.from_identifier(identifier, value, table_schema, self)
            fields.append(field)
        return fields

    def build_update_assert_fields(
        self, columns, schema_fields, lookup_name, lookup_value
    ):
        fields = []
        for eq in columns:
            if not isinstance(eq, sqlglot.expressions.EQ):
                continue
            try:
                field = DjangoField.from_eq(
                    eq, schema_fields, self, lookup_name, lookup_value
                )
            except TooComplicatedError:
                continue
            fields.append(field)
        return fields

    def build_delete_fields(self, exp, table_schema):
        if isinstance(exp, sqlglot.expressions.In):
            for literal in exp.expressions:
                yield DjangoField.from_identifier(exp.this, literal, table_schema, self)

        elif isinstance(exp, sqlglot.expressions.EQ):  # pragma: no branch
            literal = exp.expression
            yield DjangoField.from_identifier(exp.this, literal, table_schema, self)

    def build_delete_lookup_fields(self, where, table_schema):
        if isinstance(where.this, sqlglot.expressions.Paren):
            if isinstance(where.this.this, sqlglot.expressions.Or):
                left = where.this.this.left
                right = where.this.this.right
                yield list(self.build_delete_fields(left, table_schema))
                yield list(self.build_delete_fields(right, table_schema))
            elif isinstance(  # pragma: no branch
                where.this.this, sqlglot.expressions.And
            ):
                left = where.this.this.left
                right = where.this.this.right
                left_fields = self.build_delete_fields(left, table_schema)
                right_fields = self.build_delete_fields(right, table_schema)
                for fields in product(left_fields, right_fields):
                    yield list(fields)

        else:
            yield list(self.build_delete_fields(where.this, table_schema))

    def parse_columns(self, columns, row, schema_data):
        columns_by_table: Dict[str, List[DjangoField]] = {}
        for column, value in zip(columns, row):
            if not isinstance(column, sqlglot.exp.Column):
                continue

            table_schema = schema_data[column.table]
            field = DjangoField.from_column(column, value, table_schema, self)
            columns_by_table.setdefault(column.table, []).append(field)
        return columns_by_table

    def build_creates(self, query_dict, columns, row, seen_mutations, schema_data):
        inserts = []
        updates = []
        skipped_inserts = []

        columns_by_table = self.parse_columns(columns, row, schema_data)

        for table, row in columns_by_table.items():
            pk = find_pk(row)
            if pk is None:
                skipped_inserts.append(
                    SkipQuery(
                        query_dict,
                        reason=SkipQueryReason.NO_PRIMARY_KEY,
                        table=table,
                    )
                )
                continue
            elif pk.value_repr in seen_mutations.setdefault(table, set()):
                skipped_inserts.append(
                    SkipQuery(
                        query_dict,
                        reason=SkipQueryReason.SEEN_MUTATION,
                        table=table,
                    )
                )
                continue
            else:
                create, update = DjangoCreate.from_raw(
                    table, row, query_dict, schema_data, self.names
                )
                if create.primary_key.raw_value is None:
                    skipped_inserts.append(
                        SkipQuery(
                            query_dict,
                            reason=SkipQueryReason.NULL_PRIMARY_KEY,
                            table=table,
                        )
                    )
                    continue
                inserts.append(create)
                if update:
                    updates.append(update)
        return inserts, skipped_inserts, updates

    def parse_sql_queries(self, sql_queries, schema_data):
        # each select needs to become an insert
        inserts = []
        updates = []
        skipped_inserts = []
        asserts: List[AnyAssert] = []
        seen_mutations: Dict[str, Set[str]] = {}
        for query_dict in sql_queries:
            query = query_dict["query"]
            query_data = query_dict["query_data"]
            if query_data is None or query is None:
                continue

            database = query_dict["database"]
            database = DATABASES.get(database, database)
            try:
                parsed_query = sqlglot.parse_one(query, read=database)
            except sqlglot.errors.ParseError as e:  # pragma: no cover
                print("# Error parsing query:")
                print(f"# {query}")
                print(f"# {e.errors}")
                continue

            if isinstance(parsed_query, sqlglot.exp.Select):
                columns = [column for column in parsed_query.iter_expressions()]
                # Skip queries with no columns, eg .count(), .exists()
                if not any(
                    isinstance(column, sqlglot.exp.Column) for column in columns
                ):
                    alias = parsed_query.find(sqlglot.expressions.Alias)
                    count = parsed_query.find(sqlglot.expressions.Count)
                    where = parsed_query.find(sqlglot.expressions.Where)
                    if (
                        where is not None
                        and count is not None
                        and isinstance(count.this, sqlglot.expressions.Star)
                    ):
                        if query_data[0] == 0:
                            continue
                        if isinstance(where.this, sqlglot.expressions.EQ):
                            column = where.this.this
                            value = extract_value(where.this.expression)
                            new_inserts, new_skips, new_updates = self.build_creates(
                                query_dict,
                                [column],
                                [value],
                                seen_mutations,
                                schema_data,
                            )
                            inserts.extend(new_inserts)
                            updates.extend(new_updates)
                            skipped_inserts.extend(new_skips)
                        else:
                            skipped_inserts.append(
                                SkipQuery(
                                    query_dict, reason=SkipQueryReason.UNSUPPORTED_WHERE
                                )
                            )
                            continue
                    elif (
                        where is not None
                        and alias is not None
                        and query_data[0] == 1
                        and isinstance(where.this, sqlglot.expressions.EQ)
                    ):
                        column = where.this.this
                        value = extract_value(where.this.expression)
                        new_inserts, new_skips, new_updates = self.build_creates(
                            query_dict,
                            [column],
                            [value],
                            seen_mutations,
                            schema_data,
                        )
                        inserts.extend(new_inserts)
                        updates.extend(new_updates)
                        skipped_inserts.extend(new_skips)
                    else:
                        skipped_inserts.append(
                            SkipQuery(query_dict, reason=SkipQueryReason.NO_COLUMNS)
                        )
                    continue
                for batch in query_data:
                    for row in batch:
                        new_inserts, new_skips, new_updates = self.build_creates(
                            query_dict, columns, row, seen_mutations, schema_data
                        )
                        inserts.extend(new_inserts)
                        updates.extend(new_updates)
                        skipped_inserts.extend(new_skips)
            else:
                _table = parsed_query.find(sqlglot.exp.Table)
                if _table is None:
                    continue  # pragma: no cover
                table_name = _table.name
                table_schema = schema_data[table_name]
                columns = [column for column in parsed_query.iter_expressions()]
                if isinstance(parsed_query, sqlglot.expressions.Update):
                    where = parsed_query.find(sqlglot.expressions.Where)
                    if where is None:
                        continue  # pragma: no cover

                    if isinstance(where.this, sqlglot.expressions.In):
                        lookups = []
                        lookup_name = where.this.this.name
                        for literal in where.this.expressions:
                            lookups.append((lookup_name, literal.name))
                    elif isinstance(where.this, sqlglot.expressions.Paren):
                        continue
                    else:
                        lookup_name = where.this.left.name
                        lookup_value = where.this.right.name
                        lookups = [(lookup_name, lookup_value)]

                    verbose_name = table_schema["verbose_name"]
                    for lookup_name, lookup_value in lookups:
                        variable_name = self.names.next_variable_name(
                            verbose_name, lookup_name, lookup_value
                        )

                        fields = self.build_update_assert_fields(
                            columns, table_schema["fields"], lookup_name, lookup_value
                        )

                        assert_update = AssertUpdate(
                            variable_name,
                            table_schema["model_name"],
                            table_name,
                            fields,
                            query_dict,
                        )
                        asserts.append(assert_update)
                        if table_schema["fields"][lookup_name][  # pragma: no branch
                            "primary_key"
                        ]:
                            seen_mutations.setdefault(table_name, set()).add(
                                lookup_value
                            )
                elif isinstance(parsed_query, sqlglot.expressions.Delete):
                    where = parsed_query.find(sqlglot.expressions.Where)
                    if where is None:
                        continue  # pragma: no cover

                    model = table_schema["model_name"]
                    module = table_schema["model_module"]
                    for fields in self.build_delete_lookup_fields(where, table_schema):
                        assert_delete = AssertDelete(module, model, fields, query_dict)
                        asserts.append(assert_delete)
                elif isinstance(  # pragma: no branch
                    parsed_query, sqlglot.expressions.Insert
                ):
                    for assert_insert in self.build_assert_inserts(
                        parsed_query,
                        query_dict,
                        table_schema,
                        table_name,
                        seen_mutations,
                    ):
                        asserts.append(assert_insert)

        deduplicated = unique_everseen(inserts)
        sql_fixtures = sorted_inserts(deduplicated, schema_data)
        sql_fixtures.extend(updates)

        updated_creates = []
        for create in sql_fixtures:
            create.update_fields(self.names, schema_data)
            if not create.is_empty():
                updated_creates.append(create)

        asserts = merge_asserts(asserts)
        for assert_ in asserts:
            assert_.update_fields(self.names, schema_data)
        return updated_creates, asserts, skipped_inserts


def collect_gfk(fields, name, schema, names):
    try:
        from django.contrib.contenttypes.models import ContentType
    except ImportError:  # pragma: no cover
        return fields

    content_type_field = None
    object_id_field = None

    for field in fields:
        if field.schema["field_name"] == schema["gfk_content_type_field"]:
            content_type_field = field
        elif field.schema["field_name"] == schema["gfk_object_id_field"]:
            object_id_field = field

    if content_type_field is None or object_id_field is None:
        return fields  # pragma: no cover

    try:
        content_type = ContentType.objects.get(pk=content_type_field.raw_value)
    except ContentType.DoesNotExist:  # pragma: no cover
        return fields

    model = content_type.model_class()
    value = names.next_variable_name(
        verbose_name(model),
        model._meta.pk.attname,  # type: ignore
        object_id_field.raw_value,
    )
    gfk_field = DjangoField(
        name=name,
        column_name=name,
        value=value,
        schema=schema,
    )
    gfk_field.value_repr = gfk_field.value = value
    collected = []
    for field in fields:
        if field == content_type_field or field == object_id_field:
            continue
        collected.append(field)
    collected.append(gfk_field)
    return collected


def find_pk(fields):
    for field in fields:
        if field.schema["primary_key"]:
            return field
    return None  # pragma: no cover


def find_update_fields(fields, table):
    create_fields = []
    update_fields = []
    for field in fields:
        if field.schema["is_relation"] and field.schema["related_model"] == table:
            update_fields.append(field)
        else:
            create_fields.append(field)
    return create_fields, update_fields


def merge_creates(creates):
    fields = []
    seen = set()
    for create in creates:
        for field in create.fields:
            if (field.name, field.value_repr) not in seen:  # pragma: no branch
                seen.add((field.name, field.value_repr))
                fields.append(field)
    first_create = creates[0]
    return DjangoCreate(
        first_create.table,
        first_create.module,
        first_create.model,
        first_create.primary_key,
        fields,
        first_create.variable_name,
        first_create.query,
    )


def sorted_inserts(deduplicated, schema_data):
    objects: Dict[Tuple[str, str, Any], List[DjangoCreate]] = {}
    for obj in deduplicated:
        objects.setdefault(
            (obj.table, obj.primary_key.column_name, obj.primary_key.raw_value), []
        ).append(obj)
    objects = {key: merge_creates(creates) for key, creates in objects.items()}
    graph = []
    for obj in objects.values():
        schema = schema_data[obj.table]["fields"]
        keys = (
            (
                schema[field.column_name]["related_model"],
                schema[field.column_name]["related_pk"],
                field.raw_value,
            )
            for field in obj.fields
            if schema[field.column_name]["is_relation"]
        )
        graph.append((obj, [objects[key] for key in keys if key in objects]))

    sorter = TopologicalSorter(graph)
    return list(sorter.static_order())


AnyAssert = Union[AssertDelete, AssertInsert, AssertUpdate]


def merge_asserts(asserts):
    lookup: Dict[str, AnyAssert] = {}
    merged: List[AnyAssert] = []
    for _assert in asserts:
        if isinstance(_assert, AssertInsert):
            lookup[_assert.variable_name] = _assert
            merged.append(_assert)
        elif isinstance(_assert, AssertUpdate):
            if _assert.variable_name not in lookup:
                lookup[_assert.variable_name] = _assert
                merged.append(_assert)
                continue
            insert = lookup[_assert.variable_name]
            insert.fields = merge_fields(insert, _assert)
        else:
            merged.append(_assert)

    return tidy_recreates(tidy_deletes(merged))


def tidy_deletes(asserts):
    lookup: Dict[str, List[List[DjangoField]]] = {}
    tidied: List[AnyAssert] = []
    for _assert in reversed(asserts):
        if isinstance(_assert, AssertDelete):
            lookup.setdefault(_assert.model, []).append(_assert.grouped_fields())
            tidied.append(_assert)
        else:
            if _assert.model not in lookup:
                tidied.append(_assert)
                continue

            if any(
                would_be_deleted(_assert, delete_fields)
                for delete_fields in lookup[_assert.model]
            ):
                continue
            tidied.append(_assert)

    return list(reversed(tidied))


def tidy_recreates(asserts):
    lookup: Dict[str, List[List[DjangoField]]] = {}
    tidied: List[AnyAssert] = []
    for _assert in reversed(asserts):
        if isinstance(_assert, AssertUpdate):
            tidied.append(_assert)
        elif isinstance(_assert, AssertInsert):
            lookup.setdefault(_assert.model, []).append(_assert.fields)
            tidied.append(_assert)
        else:
            if _assert.model not in lookup:
                tidied.append(_assert)
                continue

            if any(
                would_be_recreated(_assert, create_fields)
                for create_fields in lookup[_assert.model]
            ):
                continue
            tidied.append(_assert)

    return list(reversed(tidied))


def would_be_deleted(_assert, delete_fields):
    if not any(field.name in delete_fields for field in _assert.fields):
        return False
    for field in _assert.fields:
        if field.name not in delete_fields:
            continue
        if field not in delete_fields[field.name]:
            return False
    return True


def would_be_recreated(delete, create_fields):
    reduced_delete_fields = []
    for delete_field in delete.fields:
        if any(delete_field == field for field in create_fields):
            continue
        reduced_delete_fields.append(delete_field)
    if not reduced_delete_fields:
        return True
    delete.fields = reduced_delete_fields
    return False


def merge_fields(insert, update):
    names = {f.name for f in update.fields}
    fields = []
    for field in insert.fields:
        if field.name in names:
            continue
        fields.append(field)
    fields.extend(update.fields)
    return fields



import json
from datetime import datetime

from more_itertools import locate, peekable


def _format_header(header: str) -> str:
    header = header.upper().replace("-", "_")
    if header in ("CONTENT_LENGTH", "CONTENT_TYPE"):
        return header
    return f"HTTP_{header}"


def get_request_headers(request):
    request_headers = {
        _format_header(header): value for header, value in request["headers"].items()
    }
    if "HTTP_COOKIE" in request_headers and request_headers["HTTP_COOKIE"] == "":
        del request_headers["HTTP_COOKIE"]

    request_headers_to_delete = [
        "CONTENT_LENGTH",
        "HTTP_HOST",
        "HTTP_X_FORWARDED_FOR",
        "HTTP_X_FORWARDED_PROTO",
        "HTTP_CONNECTION",
        "HTTP_CACHE_CONTROL",
        "HTTP_DNT",
        "HTTP_SEC_CH_UA",
        "HTTP_USER_AGENT",
        "HTTP_ACCEPT",
        "HTTP_SEC_FETCH_DEST",
        "HTTP_SEC_CH_UA_MOBILE",
        "HTTP_REFERER",
        "HTTP_ACCEPT_ENCODING",
        "HTTP_ACCEPT_LANGUAGE",
        "HTTP_SEC_FETCH_SITE",
        "HTTP_SEC_FETCH_MODE",
        "HTTP_SEC_FETCH_USER",
        "HTTP_SEC_FETCH_DEST",
        "HTTP_SEC_CH_UA_PLATFORM",
        "HTTP_ORIGIN",
        "HTTP_UPGRADE_INSECURE_REQUESTS",
    ]

    for request_header in request_headers_to_delete:
        if request_header in request_headers:
            del request_headers[request_header]

    if "CONTENT_TYPE" in request_headers:
        # This is a special header, which ends up influencing
        # how the django test client formats different parts of the
        # request. It's provided to the test client as a lowercased
        # argument
        content_type = request_headers["CONTENT_TYPE"]
        del request_headers["CONTENT_TYPE"]
        if "multipart/form-data" not in content_type:
            request_headers["content_type"] = content_type
    return request_headers


def group_frames(frames):
    request_indices = peekable(locate(frames, lambda f: f["type"] == "django_request"))
    response_indices = locate(frames, lambda f: f["type"] == "django_response")
    while True:
        try:
            request_index = next(request_indices)
            response_index = next(response_indices)
            while request_indices.peek(len(frames)) < response_index:
                request_index = next(request_indices)
            while response_index < request_index:
                response_index = next(response_indices)
        except StopIteration:
            break

        yield (
            frames[request_index],
            frames[request_index + 1 : response_index],
            frames[response_index],
        )


def parse_request_frames(frames):
    served_request_frames = []
    for request, inner_frames, response in group_frames(frames):
        served_request_frames.append(
            {
                "request": request,
                "response": response,
                "templates": [
                    frame["template"]
                    for frame in inner_frames
                    if frame["type"] == "django_template_start"
                ],
                "frames": inner_frames,
            }
        )
    return served_request_frames


def get_response_json(response):
    content_type = response.get("headers", {}).get("Content-Type", "")
    if "application/json" in content_type:
        return json.loads(response["content"])
    return None


def drop_django_templates(template_names):
    return [
        name
        for name in template_names
        if name is not None  # Templates generated from a string might not have a name
        and not name.startswith("django/")
    ]


def format_timestamp(timestamp):
    return datetime.utcfromtimestamp(timestamp).isoformat(timespec="seconds")



"""
MCP server implementation that is compatible with Python 3.8+.
This module contains the core infrastructure for the Model Context Protocol server.
"""

import asyncio
import inspect
import json
import logging
import sys
import traceback
from dataclasses import asdict, dataclass, field
from typing import Any, Callable, Dict, List, TypeVar, get_type_hints

from .version import __version__

# Set up logging
logger = logging.getLogger("kolo.mcp")
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)

# Define protocol constants
PROTOCOL_VERSION = "2024-11-05"

# Type definitions
T = TypeVar("T")
ToolFunc = Callable[..., Any]


@dataclass
class Tool:
    """Representation of a tool in the MCP protocol."""

    name: str
    func: ToolFunc
    description: str
    input_schema: Dict[str, Any]


@dataclass
class MCPServerInfo:
    """Server information for MCP protocol."""

    name: str
    version: str = ""

    def __post_init__(self):
        if not self.version:
            self.version = __version__


@dataclass
class MCPCapabilities:
    """Capabilities supported by the MCP server."""

    experimental: Dict[str, Any] = field(default_factory=dict)
    prompts: Dict[str, bool] = field(default_factory=lambda: {"listChanged": False})
    resources: Dict[str, bool] = field(
        default_factory=lambda: {"subscribe": False, "listChanged": False}
    )
    tools: Dict[str, bool] = field(default_factory=lambda: {"listChanged": False})
    roots: Dict[str, bool] = field(default_factory=lambda: {"listChanged": False})
    sampling: Dict[str, Any] = field(default_factory=dict)


class MCPServer:
    """A minimal implementation of the Model Context Protocol server compatible with Python 3.8+."""

    def __init__(self, name: str):
        self.server_info = MCPServerInfo(name=name)
        self.capabilities = MCPCapabilities()
        self.tools: Dict[str, Tool] = {}
        logger.info(f"Initialized MCP server with name: {name}")

    def tool(self) -> Callable[[ToolFunc], ToolFunc]:
        """Decorator to register a function as an MCP tool."""

        def decorator(func: ToolFunc) -> ToolFunc:
            tool_name = func.__name__
            type_hints = get_type_hints(func)
            description = func.__doc__ or ""

            # Extract parameter information
            sig = inspect.signature(func)
            properties = {}
            required = []

            for name, param in sig.parameters.items():
                if name == "self" or name == "cls":
                    continue

                param_type = type_hints.get(name, Any)
                prop: Dict[str, Any] = {}

                # Handle basic type annotations
                if param_type is inspect.Parameter.empty:
                    prop["type"] = "string"  # Default to string
                elif isinstance(param_type, type):
                    if issubclass(param_type, str):
                        prop["type"] = "string"
                    elif issubclass(param_type, int):
                        prop["type"] = "integer"
                    elif issubclass(param_type, float):
                        prop["type"] = "number"
                    elif issubclass(param_type, bool):
                        prop["type"] = "boolean"
                elif (
                    isinstance(param_type, List)
                    or getattr(param_type, "__origin__", None) is list
                ):
                    prop["type"] = "array"
                    # Create a separate dictionary for the items schema
                    # This fixes the type error where a Dict was being assigned to a str
                    item_type_dict = {"type": "string"}
                    prop["items"] = item_type_dict
                else:
                    # For complex types, we'll just use "object"
                    prop["type"] = "object"

                # Check if parameter has default (is optional)
                if param.default is inspect.Parameter.empty:
                    required.append(name)
                else:
                    # Add default value if available
                    if param.default is not None:
                        # Ensure default values have proper types in schema
                        if isinstance(param.default, bool):
                            # Make sure boolean defaults are properly typed
                            prop["default"] = param.default
                            # Force type to boolean for boolean defaults
                            prop["type"] = "boolean"
                        else:
                            prop["default"] = param.default

                properties[name] = prop

            # Create JSON schema for the tool
            input_schema = {
                "type": "object",
                "title": f"{tool_name}Arguments",
                "properties": properties,
            }

            if required:
                input_schema["required"] = required

            # Register the tool
            self.tools[tool_name] = Tool(
                name=tool_name,
                func=func,
                description=description,
                input_schema=input_schema,
            )

            logger.info(f"Registered tool: {tool_name}")
            return func

        return decorator

    async def handle_request(self, request_data: str) -> str:
        """Handle an incoming JSON-RPC request."""
        try:
            logger.debug(f"Handling request: {request_data.strip()}")

            # Parse the request data
            try:
                request = json.loads(request_data)
            except json.JSONDecodeError as e:
                logger.error(f"Failed to parse JSON: {e}")
                return self._error_response(None, -32700, f"Parse error: {str(e)}")

            # Extract JSON-RPC fields
            jsonrpc = request.get("jsonrpc")
            method = request.get("method")
            params = request.get("params", {})
            request_id = request.get("id")

            logger.info(f"Processing method: {method} with id: {request_id}")

            if jsonrpc != "2.0":
                logger.warning(f"Invalid JSON-RPC version: {jsonrpc}")
                return self._error_response(
                    request_id, -32600, "Invalid Request: expected jsonrpc 2.0"
                )

            if method == "initialize":
                response = self._initialize_response(request_id)
                logger.debug(f"Initialize response JSON: {response}")
                return response
            elif method == "tools/list":
                response = self._tools_list_response(request_id)
                logger.debug(f"Tools list response JSON: {response}")
                return response
            elif method == "roots/list":
                # Handle roots/list request
                response = self._roots_list_response(request_id)
                logger.debug(f"Roots list response JSON: {response}")
                return response
            elif method == "tools/call":
                # Handle tools/call method (used by Anthropics MCP tool inspector)
                logger.info("Processing tools/call method")
                tool_name = params.get("name")
                tool_params = params.get(
                    "arguments", {}
                )  # Changed from parameters to arguments

                if not tool_name:
                    logger.warning("Tool name not provided in tools/call")
                    return self._error_response(
                        request_id, -32602, "Invalid params: tool name not provided"
                    )

                if tool_name not in self.tools:
                    logger.warning(f"Tool not found: {tool_name}")
                    return self._error_response(
                        request_id, -32601, f"Tool '{tool_name}' not found"
                    )

                logger.info(f"Executing tool via tools/call: {tool_name}")
                response = await self._execute_tool(request_id, tool_name, tool_params)
                logger.debug(f"Tool execution response JSON: {response}")
                return response
            elif method.startswith("notifications/"):
                # Handle notifications (like cancelled requests)
                notification_type = method.split("/")[1]
                logger.info(f"Received notification: {notification_type}")

                # No response needed for notifications as per JSON-RPC spec
                # Just acknowledge with an empty success response
                return json.dumps({"jsonrpc": "2.0", "id": request_id, "result": {}})
            elif method.startswith("tools/execute/"):
                tool_name = method.split("/")[-1]
                if tool_name not in self.tools:
                    logger.warning(f"Tool not found: {tool_name}")
                    return self._error_response(
                        request_id, -32601, f"Tool '{tool_name}' not found"
                    )

                logger.info(f"Executing tool: {tool_name}")
                response = await self._execute_tool(request_id, tool_name, params)
                logger.debug(f"Tool execution response JSON: {response}")
                return response
            else:
                logger.warning(f"Method not found: {method}")
                return self._error_response(
                    request_id, -32601, f"Method '{method}' not found"
                )

        except json.JSONDecodeError:
            logger.error("JSON parse error")
            return self._error_response(None, -32700, "Parse error")
        except Exception as e:
            logger.error(f"Internal error: {str(e)}")
            logger.error(traceback.format_exc())
            return self._error_response(None, -32603, f"Internal error: {str(e)}")

    def _initialize_response(self, request_id: Any) -> str:
        """Generate the initialize response with server info and capabilities."""
        response = {
            "jsonrpc": "2.0",
            "id": request_id,
            "result": {
                "protocolVersion": PROTOCOL_VERSION,
                "capabilities": asdict(self.capabilities),
                "serverInfo": asdict(self.server_info),
            },
        }
        logger.debug(f"Initialize response: {response}")
        logger.debug(f"Server info: {asdict(self.server_info)}")
        logger.debug(f"Capabilities: {asdict(self.capabilities)}")
        return json.dumps(response)

    def _tools_list_response(self, request_id: Any) -> str:
        """Generate the tools/list response."""
        tools_list = []

        for name, tool in self.tools.items():
            tools_list.append(
                {
                    "name": name,
                    "description": tool.description,
                    "inputSchema": tool.input_schema,
                }
            )

        response = {"jsonrpc": "2.0", "id": request_id, "result": {"tools": tools_list}}
        logger.debug(f"Tools list response contains {len(tools_list)} tools")
        return json.dumps(response)

    def _roots_list_response(self, request_id: Any) -> str:
        """Generate the roots/list response."""
        roots_list = [
            {
                "id": "kolo",
                "name": "Kolo MCP Tools",
                "description": "Tools for deep Python tracing and debugging",
            }
        ]

        response = {"jsonrpc": "2.0", "id": request_id, "result": {"roots": roots_list}}
        logger.debug(f"Roots list response contains {len(roots_list)} roots")
        return json.dumps(response)

    async def _execute_tool(
        self, request_id: Any, tool_name: str, params: Dict[str, Any]
    ) -> str:
        """Execute a tool and return the response."""
        tool = self.tools.get(tool_name)
        if not tool:
            logger.warning(f"Tool not found for execution: {tool_name}")
            return self._error_response(
                request_id, -32601, f"Tool '{tool_name}' not found"
            )

        try:
            # Call the tool function with the provided parameters
            logger.info(f"Calling tool function: {tool_name}")
            logger.debug(f"Tool params: {params}")
            result = tool.func(**params)

            # Handle async functions
            if inspect.iscoroutine(result):
                logger.debug(f"Awaiting coroutine result for tool: {tool_name}")
                result = await result

            logger.info(f"Tool {tool_name} result: {result}")

            # Format the response according to MCP protocol
            if isinstance(result, List):
                # For lists, we need to format each item as a separate line item
                content_items = []
                for item in result:
                    content_items.append({"type": "text", "text": str(item)})

                response = {
                    "jsonrpc": "2.0",
                    "id": request_id,
                    "result": {"content": content_items},
                }
            else:
                # For non-list results, use the standard format
                response = {
                    "jsonrpc": "2.0",
                    "id": request_id,
                    "result": {
                        "content": [
                            {
                                "type": "text",
                                "text": str(result)
                                if isinstance(result, (str, int, float, bool))
                                else json.dumps(result),
                            }
                        ]
                    },
                }

            logger.info(f"Tool {tool_name} executed successfully")
            return json.dumps(response)

        except Exception as e:
            error_msg = f"Error executing tool {tool_name}: {str(e)}"
            logger.error(error_msg)
            logger.error(traceback.format_exc())
            return self._error_response(request_id, -32603, error_msg)

    def _error_response(self, request_id: Any, code: int, message: str) -> str:
        """Generate an error response."""
        response = {
            "jsonrpc": "2.0",
            "id": request_id,
            "error": {"code": code, "message": message},
        }
        logger.debug(f"Error response: {response}")
        return json.dumps(response)

    async def read_stdin_line(self) -> str:
        """Read a line from stdin in an asynchronous way."""
        line = await asyncio.get_event_loop().run_in_executor(None, sys.stdin.readline)
        logger.debug(f"Read line from stdin: {line.strip()}")
        return line

    async def run_stdio(self):
        """Run the server reading from stdin and writing to stdout."""
        logger.info("Starting MCP server on stdio transport")
        try:
            # Simpler approach based on the FastMCP implementation pattern
            while True:
                line = await self.read_stdin_line()
                if not line:
                    logger.info("Received empty line, exiting")
                    break

                if line.strip():
                    try:
                        if line.startswith("{"):
                            # Direct JSON-RPC request
                            request_data = line
                        else:
                            # Header-based request
                            if line.startswith("Content-Length:"):
                                content_length = int(line.strip().split(": ")[1])
                                # Skip empty line
                                await self.read_stdin_line()
                                # Read the content
                                request_data = (
                                    await asyncio.get_event_loop().run_in_executor(
                                        None,
                                        lambda: sys.stdin.buffer.read(
                                            content_length
                                        ).decode("utf-8"),
                                    )
                                )
                            else:
                                logger.warning(f"Invalid header: {line.strip()}")
                                continue

                        # Handle the request
                        response = await self.handle_request(request_data)

                        # Log the response for debugging
                        logger.debug(f"Response: {response}")

                        # Write response - always with newline for MCP tool inspector
                        sys.stdout.write(response)
                        sys.stdout.write("\n")
                        sys.stdout.flush()
                    except Exception as e:
                        logger.error(f"Error handling request: {e}")
                        logger.error(traceback.format_exc())

                        # Send error response
                        error_response = json.dumps(
                            {
                                "jsonrpc": "2.0",
                                "error": {
                                    "code": -32603,
                                    "message": f"Internal error: {str(e)}",
                                },
                                "id": None,
                            }
                        )
                        sys.stdout.write(error_response)
                        sys.stdout.write("\n")
                        sys.stdout.flush()
        except Exception as e:
            logger.error(f"Fatal error in MCP server loop: {e}")
            logger.error(traceback.format_exc())



from typing import Optional, Tuple

from .db import db_connection, get_db_path


def create_kolo_kv_table(connection) -> None:
    create_table_query = """
    CREATE TABLE IF NOT EXISTS kolo_kv (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        key TEXT UNIQUE NOT NULL,
        value TEXT,
        updated_at TEXT NOT NULL DEFAULT (strftime('%Y-%m-%dT%H:%M:%fZ', 'now')),
        created_at TEXT NOT NULL DEFAULT (strftime('%Y-%m-%dT%H:%M:%fZ', 'now'))
    );
    """

    connection.execute(create_table_query)


def set(key, value) -> None:
    db_path = get_db_path()
    with db_connection(db_path) as connection:
        create_kolo_kv_table(connection)

        cursor = connection.execute("SELECT id FROM kolo_kv WHERE key = ?;", (key,))

        if cursor.fetchone():
            connection.execute(
                "UPDATE kolo_kv SET value = ?, updated_at = (strftime('%Y-%m-%dT%H:%M:%fZ', 'now')) WHERE key = ?;",
                (value, key),
            )
        else:
            connection.execute(
                "INSERT INTO kolo_kv (key, value) VALUES (?, ?);", (key, value)
            )


def get(key) -> Tuple[int, str, str, str, str]:
    db_path = get_db_path()
    with db_connection(db_path) as connection:
        create_kolo_kv_table(connection)

        cursor = connection.execute(
            "SELECT id, key, value, updated_at, created_at FROM kolo_kv WHERE key = ?;",
            (key,),
        )
        result = cursor.fetchone()
        if not result:
            raise KeyError(f"Key {key} not found in kolo_kv")

        return result


def get_value(key) -> Optional[str]:
    id, key, value, updated_at, created_at = get(key)

    return value



from datetime import datetime
from itertools import chain
from typing import List

from ..db import SchemaNotFoundError, load_schema_for_commit_sha
from ..django_schema import get_schema_for_commit
from ..git import COMMIT_SHA
from .factories import build_factories, import_factories
from .loaders import import_string
from .outbound import parse_outbound_frames
from .queries import DjangoCreate, SQLParser
from .request import (
    drop_django_templates,
    format_timestamp,
    get_request_headers,
    get_response_json,
    parse_request_frames,
)


def process_django_schema(context):
    traces = context["_traces"]
    use_saved_schemas = context["_use_saved_schemas"]
    for data in traces.values():
        commit_sha = data["trace"]["current_commit_sha"]
        db_path = context["_db_path"]
        data["schema_data"] = get_schema_for_commit(
            commit_sha, db_path, use_saved_schemas
        )


def reused_fixture(new_fixture, fixtures):
    for fixture in fixtures:
        if (
            fixture.table == new_fixture.table
            and fixture.model == new_fixture.model
            and fixture.primary_key == new_fixture.primary_key
        ):
            return True
    return False


def process_sql_queries(context):
    parser = SQLParser(context["_config"])

    all_fixtures: List[DjangoCreate] = []
    for section in context["sections"]:
        frames = section["frames"]
        sql_queries = [frame for frame in frames if frame["type"] == "end_sql_query"]
        sql_fixtures, asserts, skipped_inserts = parser.parse_sql_queries(
            sql_queries, section["schema_data"]
        )
        sql_fixtures = [f for f in sql_fixtures if not reused_fixture(f, all_fixtures)]
        all_fixtures.extend(sql_fixtures)
        section["asserts"] = asserts
        section["sql_fixtures"] = sql_fixtures
        section["skipped_inserts"] = skipped_inserts


def process_factories(context):
    for section in context["sections"]:
        sql_fixtures = section.get("sql_fixtures", [])
        factory_configs = import_factories(context["_config"])
        factories = build_factories(sql_fixtures, factory_configs)
        section["sql_fixtures"] = factories


def process_django_request(context):
    sections = []

    for data in context["_traces"].values():
        frames = data["frames"]
        served_request_frames = parse_request_frames(frames)

        if not served_request_frames:
            sections.append({"frames": frames, "schema_data": data["schema_data"]})

        for served in served_request_frames:
            request = served["request"]
            response = served["response"]
            template_names = drop_django_templates(served["templates"])

            sections.append(
                {
                    "frames": served["frames"],
                    "request": request,
                    "request_headers": get_request_headers(request),
                    "request_timestamp": format_timestamp(request["timestamp"]),
                    "response": response,
                    "response_json": get_response_json(response),
                    "schema_data": data["schema_data"],
                    "template_names": template_names,
                }
            )

    return {"sections": sections}


def process_outbound_requests(context):
    for section in context["sections"]:
        frames = section["frames"]
        outbound_request_frames = parse_outbound_frames(frames)
        section["outbound_request_frames"] = outbound_request_frames


def process_django_version(context):
    try:
        from django import __version__ as django_version
    except ImportError:  # pragma: no cover
        django_version = ""
    return {"django_version": django_version}


def process_time_travel(context):
    """Choose time_machine unless only freezegun is installed"""
    try:
        import time_machine
    except ImportError:
        try:
            import freezegun
        except ImportError:  # pragma: no cover
            pass
        else:
            return {
                "time_travel_import": "import freezegun",
                "time_travel_call": "freezegun.freeze_time",
                "time_travel_tick": ", tick=True",
            }
    return {
        "time_travel_import": "import time_machine",
        "time_travel_call": "time_machine.travel",
        "time_travel_tick": "",
    }


def load_processors(config):
    default_processors = (
        process_django_version,
        process_django_schema,
        process_django_request,
        process_outbound_requests,
        process_sql_queries,
        process_factories,
        process_time_travel,
    )
    custom_processors = config.get("test_generation", {}).get("trace_processors", [])
    custom_processors = map(import_string, custom_processors)
    return tuple(chain(default_processors, custom_processors))


def run_processors(processors, context):
    for processor in processors:
        processor_output = processor(context)
        if processor_output:
            context.update(processor_output)



from __future__ import annotations

import traceback
import types
from itertools import starmap
from typing import Dict, TypedDict

try:
    from django.db.models import Model as DjangoModel
except ImportError:  # pragma: no cover

    class DjangoModel:  # type: ignore
        """Stub type so isinstance returns False"""


from ..serialize import frame_path
from .core import get_ignore_frames, get_include_frames, library_filter


class ExceptionFrameInfo(TypedDict):
    path: str
    co_name: str
    locals: Dict[str, object]
    expanded_locals: Dict[str, Dict[str, object]]


def serialize_exception_frame(frame, expanded_locals) -> "ExceptionFrameInfo":
    return {
        "path": frame_path(frame),
        "co_name": frame.f_code.co_name,
        "locals": frame.f_locals,
        "expanded_locals": expanded_locals,
    }


def include_match(frame: types.FrameType, context) -> bool:
    for include_filter in context["include_frames"]:
        if include_filter(frame, event="return", arg=None):
            return True
    return False


def ignore_match(frame: types.FrameType, context) -> bool:
    for ignore_filter in context["ignore_frames"]:
        if ignore_filter(frame, event="return", arg=None):
            return True
    return False


def process_exception(
    frame: types.FrameType,
    event: str,
    arg: object,
    context,
):
    frame_locals = frame.f_locals
    exc_type, exc_value, exc_traceback = frame_locals["exc_info"]

    recorded_exception_frames = []
    expanded_locals_for_frames = []
    frames = (frame[0] for frame in traceback.walk_tb(exc_traceback))
    for frame in frames:
        _include_match = include_match(frame, context)
        if not _include_match and ignore_match(frame, context):
            continue

        if _include_match or not library_filter(frame):
            frame_locals = frame.f_locals

            expanded_locals = {}
            for key, value in frame_locals.items():
                if hasattr(value, "__dict__") and isinstance(value, DjangoModel):
                    expanded_locals[key] = vars(value)

            recorded_exception_frames.append(frame)
            expanded_locals_for_frames.append(expanded_locals)

    exception_with_traceback = traceback.format_exception(
        exc_type, exc_value, exc_traceback
    )

    zipped_frames = zip(recorded_exception_frames, expanded_locals_for_frames)
    exception_frames = list(starmap(serialize_exception_frame, zipped_frames))
    bottom_exception_frame = exception_frames[-1] if exception_frames else None

    return {
        "exception_summary": traceback.format_exception_only(exc_type, exc_value),
        "exception_with_traceback": exception_with_traceback,
        "exception_frames": exception_frames,
        "bottom_exception_frame": bottom_exception_frame,
    }


def build_context(config):
    return {
        "include_frames": get_include_frames(config),
        "ignore_frames": get_ignore_frames(config),
    }


exception = {
    "co_names": ("handle_uncaught_exception",),
    "path_fragment": "django",
    "events": ["call"],
    "call_type": "exception",
    "return_type": "exception",
    "process": process_exception,
    "build_context": build_context,
}



from __future__ import annotations

import asyncio
import gzip
import json
import os
import sqlite3
import subprocess
import sys
import time
from datetime import datetime, timedelta
from http.server import ThreadingHTTPServer
from io import BytesIO
from pathlib import Path
from runpy import run_module, run_path
from typing import Any, Dict, List, Optional, Sequence, Union

import click
import httpx
import msgpack

from . import django_schema
from .cli_mcp_shared import get_compact_traces, get_formatted_traces, get_node_data
from .config import load_config
from .db import (
    TraceNotFoundError,
    convert_json_to_msgpack,
    create_schema_table,
    db_connection,
    delete_traces_before,
    delete_traces_by_id,
    list_traces_from_db,
    load_trace_from_db,
    pin_trace,
    save_schema,
    save_trace_in_sqlite,
    setup_db,
    unpin_trace,
    vacuum_db,
)
from .generate_tests import generate_from_trace_ids, generate_test_intermediate_format
from .profiler import enable
from .serialize import load_msgpack, monkeypatch_queryset_repr
from .upload import upload_to_dashboard
from .utils import (
    extract_main_frames_from_data,
    get_terminal_formatter,
    highlight_python,
    highlight_sql,
    maybe_format,
)
from .web.server import KoloRequestHandler

DATETIME_FORMATS = click.DateTime(
    (
        "%Y-%m-%d",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%dT%H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%S.%f",
    )
)


DJANGO_SETTINGS_ERROR = """Django settings not found.
Use `--settings` or set the `DJANGO_SETTINGS_MODULE` environment variable."""

TRACE_NOT_FOUND_ERROR = "Could not find trace_id: `{trace_id}`"


@click.group()
def cli():
    # Ensure the current working directory is on the path.
    # Important when running the `kolo` script installed by setuptools.
    # Not really necessary when using `python -m kolo`, but doesn't hurt.
    # Note, we use 1, not 0: https://stackoverflow.com/q/10095037
    # This probably doesn't matter for our use case, but it doesn't hurt.
    sys.path.insert(1, ".")


def python_noop_profiler(frame, event, arg):  # pragma: no cover
    pass


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("path")
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.option(
    "--one-trace-per-test",
    default=False,
    is_flag=True,
    help="Generate a trace for each test traced by Kolo.",
)
@click.option("--noop", default=False, is_flag=True, hidden=True)
@click.option(
    "--use-monitoring",
    default=False,
    is_flag=True,
    help="Enable Kolo implementation using sys.monitoring",
)
def run(path, args, one_trace_per_test, noop, use_monitoring):
    """
    Profile Python code using Kolo.

    PATH is the path to the python file or module being profiled.
    """
    if path == "python":
        path, *args = args
        if path == "-m":
            path, *args = args
            module = True
        else:
            module = False
    elif path.endswith(".py"):
        module = False
    else:
        module = True

    existing_profiler = sys.getprofile()
    if existing_profiler:
        raise click.ClickException(
            f"Cannot run Kolo: {existing_profiler} is already active."
        )

    # Monkeypatch sys.argv, so the profiled code doesn't get confused
    # Without this, the profiled code would see extra args it doesn't
    # know how to handle.
    sys.argv = [path, *args]

    if noop:  # pragma: no cover
        config = load_config()
        if config.get("use_rust", True):
            from ._kolo import register_noop_profiler

            register_noop_profiler()
        else:
            sys.setprofile(python_noop_profiler)

        try:
            if module:
                run_module(path, run_name="__main__", alter_sys=True)
            else:
                run_path(path, run_name="__main__")
        finally:
            sys.setprofile(None)
        return

    if use_monitoring:
        config = {"use_monitoring": True}
    else:
        config = {}
    monkeypatch_queryset_repr()
    with enable(config, source="kolo run", _one_trace_per_test=one_trace_per_test):
        if module:
            run_module(path, run_name="__main__", alter_sys=True)
        else:
            run_path(path, run_name="__main__")


@cli.command()
@click.option(
    "--port",
    default=5656,
    help="Custom port for the server to run on. Defaults to 5656.",
)
@click.option(
    "--ip",
    default="127.0.0.1",
    help="Custom ip for the server to run on. Defaults to 127.0.0.1",
)
def server(port, ip):
    """
    Start a server to view traces.
    """
    server = ThreadingHTTPServer((ip, port), KoloRequestHandler)
    click.echo(f"View Kolo traces at http://{ip}:{port}/_kolo/")
    server.serve_forever()


@cli.group()
def trace():
    """
    Subcommands for working with Kolo traces.
    """


@trace.command()
@click.argument("path")
@click.option(
    "--created-at",
    help="Mark this trace as created at this time.",
    type=DATETIME_FORMATS,
)
def load(path, created_at):
    """
    Load a trace from a file into the Kolo database.
    """
    db_path = setup_db()

    try:
        with open(path, "rb") as dump:
            raw_data = dump.read()
    except FileNotFoundError:
        raise click.ClickException(f'File "{path}" not found')

    try:
        data = msgpack.unpackb(raw_data, strict_map_key=False)
    except Exception:
        raise click.ClickException("Trace file is not valid msgpack")

    try:
        trace_id = data["trace_id"]
    except KeyError:
        raise click.ClickException("Trace file is missing the `trace_id`")

    try:
        save_trace_in_sqlite(
            db_path,
            trace_id,
            msgpack=raw_data,
            ignore_errors=False,
            created_at=created_at,
        )
    except sqlite3.IntegrityError:
        raise click.ClickException(f"Trace ID `{trace_id}` already exists")

    click.echo(f"Loaded trace {trace_id}")


@trace.command()
@click.argument("trace_id")
@click.option("--file", help="The name of the file to save the trace to.")
@click.option(
    "--as-python",
    default=False,
    is_flag=True,
    help="Show the trace as readable Python types.",
)
@click.option(
    "--syntax-highlight",
    help="Highlight Python syntax. Ignored unless `--as-python` is set.",
    default="off",
    flag_value="dark",
    is_flag=False,
    type=click.Choice(["off", "light", "dark"], case_sensitive=False),
)
def dump(trace_id, file, as_python, syntax_highlight):
    """
    Dump a trace from the Kolo database to stdout or a specified file.
    """
    db_path = setup_db()

    try:
        msgpack_data, _ = load_trace_from_db(db_path, trace_id)
    except TraceNotFoundError:
        raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))

    if as_python:
        data = load_msgpack(msgpack_data)
        data = repr(data)
        data = maybe_format(data)
        if syntax_highlight != "off":
            formatter = get_terminal_formatter(syntax_highlight)
            data = highlight_python(data, formatter)
        if file:
            with open(file, "w") as f:
                f.write(data)
        else:
            click.echo(data)
    elif file:
        with open(file, "wb") as f:
            f.write(msgpack_data)

    else:
        click.echo(msgpack_data, nl=False)


@trace.command()
@click.option("--count", help="The number of rows to show.", default=500)
@click.option(
    "--reverse",
    help="Reverse the order of the rows: newest at the bottom.",
    default=False,
    is_flag=True,
)
def list(count, reverse):
    """
    Concisely list available traces, ordered by most recent first.

    This is useful for getting an overview of what traces are available and
    finding specific traces you might want to analyze further.

    After listing the trace, you can use `kolo trace compact <trace_id> to get more
    details about the trace.

    If you're an LLM, use `kolo trace list` often to orient yourself.
    """
    db_path = setup_db()
    found_any = False

    for formatted_trace in get_formatted_traces(db_path, count=count, reverse=reverse):
        found_any = True
        click.echo(formatted_trace)

    if not found_any:
        click.echo("No traces found")


@trace.command()
@click.argument("trace_id")
@click.option(
    "--syntax-highlight",
    help="Highlight SQL query syntax.",
    default="off",
    flag_value="dark",
    is_flag=False,
    type=click.Choice(["off", "light", "dark"], case_sensitive=False),
)
def list_queries(trace_id, syntax_highlight):
    """List all SQL queries in a trace."""
    db_path = setup_db()

    try:
        msgpack_data, _ = load_trace_from_db(db_path, trace_id)
    except TraceNotFoundError:
        raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))

    formatter = get_terminal_formatter(syntax_highlight)

    data = load_msgpack(msgpack_data)
    frames = extract_main_frames_from_data(data)
    for frame in frames:
        if frame["type"] == "end_sql_query":
            query = highlight_sql(
                frame["query"], dialect=frame["database"], formatter=formatter
            )
            click.echo(query)


@trace.command()
@click.argument("trace_id")
@click.argument("node_index", type=int)
def node(trace_id: str, node_index: int):
    """Get detailed information about a specific node in a trace.
    This is useful when you need to deeply understand what happened at a specific
    point in the execution, like examining function arguments, local variables,
    or the exact line of code being executed.
    """
    db_path = setup_db()

    async def _get_node():
        try:
            trace_data, _ = load_trace_from_db(db_path, trace_id)
            node_data = await get_node_data(trace_id, node_index, trace_data)
            click.echo(json.dumps(node_data, indent=2))
        except TraceNotFoundError:
            raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))
        except httpx.HTTPError as e:
            if hasattr(e, "response") and e.response.status_code == 404:
                raise click.ClickException(
                    f"Node {node_index} not found in trace {trace_id}"
                )
            raise click.ClickException(f"Error fetching node data: {e}")

    asyncio.run(_get_node())


@trace.command()
@click.argument("trace_ids", required=False, nargs=-1)
@click.option("--old", is_flag=True, default=False, help="Delete old traces.")
@click.option(
    "--before",
    help="Delete traces older than this datetime. Must be used with `--old`.",
    type=DATETIME_FORMATS,
)
@click.option(
    "--vacuum",
    help="Recover disk space from the Kolo database.",
    default=False,
    is_flag=True,
)
def delete(trace_ids, old, before, vacuum):
    """
    Delete one or more traces stored by Kolo.
    """

    if before is not None and old is False:
        raise click.ClickException("--before requires --old")

    if old is False and not trace_ids and vacuum is False:
        raise click.ClickException("Must specify either TRACE_IDS, --old, or --vacuum")

    if trace_ids and old:
        raise click.ClickException("Cannot specify TRACE_IDS and --old together")

    db_path = setup_db()

    if trace_ids:
        delete_traces_by_id(db_path, trace_ids)
    elif old:
        if before is None:
            before = datetime.now() - timedelta(days=30)

        deleted_count = delete_traces_before(db_path, before)
        click.echo(f"Deleted {deleted_count} old traces created before {before}.")

    if vacuum:
        vacuum_db(db_path)


@trace.command()
@click.argument("trace_id")
def pin(trace_id):
    """Pin a trace.
    This includes the trace in the output of using `kolo trace compact --pinned`,
    which can be helpful to give an LLM an overview of the key transactions in your project.
    """
    db_path = setup_db()
    try:
        if pin_trace(db_path, trace_id):
            click.echo(f"Pinned trace {trace_id}")
        else:
            raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))
    except TraceNotFoundError:
        raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))


@trace.command()
@click.argument("trace_id")
def unpin(trace_id):
    """Unpin a previously pinned trace."""
    db_path = setup_db()
    try:
        if unpin_trace(db_path, trace_id):
            click.echo(f"Unpinned trace {trace_id}")
        else:
            raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))
    except TraceNotFoundError:
        raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))


@trace.command()
@click.argument("trace_id", required=False)
@click.option(
    "--pinned", is_flag=True, help="Show compact representation of all pinned traces"
)
@click.option(
    "--returns", is_flag=True, help="Include return values in compact representation (warning: can be verbose)"
)
@click.option(
    "--recent",
    type=int,
    is_flag=False,
    flag_value=5,
    help="Show compact representation of the N most recent traces (default: 5)",
)
def compact(trace_id: str | None, pinned: bool, returns: bool, recent: int | None):
    """Get a compact representation of a specific trace.

    Get a concise yet detailed overview of everything that happened in the trace.
    You will see a tree representation of all function calls (and optionally return values),
    as well as other relevant information and points of interest in the trace like logs or sql queries.
    """
    if not any([pinned, trace_id is not None, recent is not None]):
        raise click.UsageError("Please provide a trace_id, --pinned, or --recent")

    db_path = setup_db()

    async def _get_compact():
        try:
            results = await get_compact_traces(
                db_path, trace_id, pinned=pinned, returns=returns, recent=recent or 0
            )
            for tid, compact_repr in results:
                if pinned or recent is not None:
                    click.echo(f"\n=== Trace {tid} ===")
                click.echo(compact_repr)
        except TraceNotFoundError as e:
            raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=e.args[0]))

    asyncio.run(_get_compact())


def manage_py_settings():
    import ast

    try:
        with open("manage.py") as f:
            data = f.read()
    except OSError:  # pragma: no cover
        return None
    source = ast.parse(data, "manage.py")
    for node in ast.walk(source):
        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and node.func.attr == "setdefault"
            and isinstance(node.args[0], ast.Constant)
            and node.args[0].value == "DJANGO_SETTINGS_MODULE"
            and isinstance(node.args[1], ast.Constant)
        ):
            return node.args[1].value  # pragma: no cover
    return None


def load_django(settings):
    import django

    if settings:
        os.environ["DJANGO_SETTINGS_MODULE"] = settings
    elif os.environ.get("DJANGO_SETTINGS_MODULE"):
        pass
    else:
        settings = manage_py_settings()
        if settings:
            os.environ["DJANGO_SETTINGS_MODULE"] = settings  # pragma: no cover
        else:
            raise click.ClickException(DJANGO_SETTINGS_ERROR)

    django.setup()


@cli.command()
@click.argument("trace_ids", nargs=-1)
@click.option(
    "--test-name", default="test_my_view", help="The name of the generated test."
)
@click.option(
    "--test-class",
    default="MyTestCase",
    help="The name of the generated TestCase class.",
)
@click.option(
    "--file",
    type=click.File("w"),
    help="Write the generated test to this file.",
)
@click.option(
    "--settings", default="", help="The dotted path to a Django settings module."
)
@click.option(
    "--use-saved-schemas",
    default=False,
    is_flag=True,
    help="Load Django schemas saved with `kolo store-django-model-schema` instead of using the current schema. This may be useful when generating a test from an old trace.",
)
@click.option(
    "--intermediate-format",
    default=False,
    is_flag=True,
    help="Show an intermediate format of the data Kolo has extracted from the trace for use in the test.",
)
@click.option(
    "--and-run",
    default=False,
    is_flag=True,
    help="[Experimental, Use not recommended] Immediately run the newly generated test.",
)
@click.option(
    "--unittest", "pytest", flag_value=False, help="Generate a unittest style test."
)
@click.option(
    "--pytest",
    "pytest",
    flag_value=True,
    default=True,
    help="Generate a pytest style test (default).",
)
def generate_test(
    trace_ids,
    test_name,
    test_class,
    file,
    settings,
    use_saved_schemas,
    intermediate_format,
    and_run,
    pytest,
):
    """Generate a test from a Kolo trace."""
    import logging

    logging.disable()

    try:
        load_django(settings)

        if intermediate_format:
            for data in generate_test_intermediate_format(
                *trace_ids,
                test_class=test_class,
                test_name=test_name,
                use_saved_schemas=use_saved_schemas,
            ):
                click.echo(data)
            return
        try:
            test_code = generate_from_trace_ids(
                *trace_ids,
                test_class=test_class,
                test_name=test_name,
                use_saved_schemas=use_saved_schemas,
                pytest=pytest,
            )
        except TraceNotFoundError as e:
            raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=e.args[0]))

        if and_run:
            import unittest

            from django.test.runner import DiscoverRunner

            our_globals: Dict[Any, Any] = {}
            exec(test_code, our_globals)

            # Create TestSuite manually and add test cases
            test_suite = unittest.TestSuite()
            test_suite.addTest(
                unittest.defaultTestLoader.loadTestsFromTestCase(
                    our_globals[test_class]
                )
            )

            class CustomRunner(DiscoverRunner):
                def build_suite(
                    self, test_labels: Optional[Sequence[str]] = None, **kwargs
                ) -> unittest.TestSuite:
                    return test_suite

                def suite_result(self, suite, result, **kwargs):
                    return suite, result

            runner = CustomRunner()

            # Test output will be output in the terminal as usual
            suite, result = runner.run_tests([])  # type: ignore # we're now returning a tuple (see CustomRunner.suite_result above)
        elif file:
            file.write(test_code)
        else:
            click.echo(test_code)
    finally:
        logging.disable(logging.NOTSET)


@cli.command()
@click.option(
    "--settings", default="", help="The dotted path to a Django settings module."
)
def store_django_model_schema(settings):
    """Store Django model info for test generation."""
    from .git import COMMIT_SHA

    load_django(settings)

    schema = django_schema.get_schema()

    db_path = setup_db()
    with db_connection(db_path) as connection:
        create_schema_table(connection)
        save_schema(connection, schema, COMMIT_SHA)


@cli.command()
def dbshell():  # pragma: no cover
    """
    Open a sqlite3 shell to the Kolo database.
    """
    db_path = setup_db()
    subprocess.run(["sqlite3", db_path], check=True)


@trace.command()
@click.argument("trace_id")
def upload(trace_id):
    """
    Upload a trace to the Kolo dashboard
    """
    db_path = setup_db()

    try:
        msgpack_data, _ = load_trace_from_db(db_path, trace_id)
    except TraceNotFoundError:
        raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))

    response = upload_to_dashboard(msgpack_data)

    if response.status_code == 201:
        click.echo(f"{trace_id} uploaded successfully!")
    else:
        errors = response.json()
        raise click.ClickException(errors)


@trace.command()
@click.argument("trace_id")
def download(trace_id):
    """
    Download a trace from the Kolo dashboard
    """
    db_path = setup_db()

    # TODO(later): The ability to download will ultimately likely be guarded by authenticating
    # against a given project/organisation etc so that you cannot download someone else's trace
    # just by guessing a ULID, so we'll probably need to pass those along eventually too...
    base_url = os.environ.get("KOLO_BASE_URL", "https://my.kolo.app")
    url = f"{base_url}/api/traces/{trace_id}/download"
    response = httpx.get(url)

    if response.status_code == 404:
        raise click.ClickException(f"`{trace_id}` was not found by the server.")
    elif response.status_code != 200:
        raise click.ClickException(f"Unexpected status code: {response.status_code}.")

    raw_data = response.content

    try:
        msgpack.unpackb(raw_data, strict_map_key=False)
    except Exception:
        raise click.ClickException("Downloaded trace was not valid msgpack.")

    try:
        save_trace_in_sqlite(
            db_path,
            trace_id,
            msgpack=raw_data,
            ignore_errors=False,
        )
    except sqlite3.IntegrityError:
        raise click.ClickException(f"`{trace_id}` already exists.")

    click.echo(f"`{trace_id}` downloaded successfully!")


@trace.command()
def json_to_msgpack():  # pragma: no cover
    """
    Convert all legacy json traces to msgpack
    """
    db_path = setup_db()
    count = convert_json_to_msgpack(db_path)
    click.echo(f"{count} traces converted!")


@cli.group()
def ci():
    """
    Subcommands for CI-related operations.
    """
    pass


def compress_trace(trace_data: bytes) -> bytes:
    compressed_data = BytesIO()
    with gzip.GzipFile(fileobj=compressed_data, mode="wb") as gz:
        gz.write(trace_data)
    return compressed_data.getvalue()


def create_trace_group() -> str:
    auth_token = os.environ["KOLO_TOKEN"]
    headers = {"Authorization": f"Bearer {auth_token}"}
    base_url = os.environ.get("KOLO_BASE_URL", "https://my.kolo.app")
    response = httpx.post(
        f"{base_url}/api/trace-groups", headers=headers, json={"name": "ci"}
    )
    if response.status_code != 201:
        error_message = f"Failed to create trace group. Status code: {response.status_code}\n{response.text}"

        click.echo(click.style(error_message, fg="red"), err=True)
        raise click.Abort()

    trace_group = response.json()

    return trace_group["id"]


async def upload_trace(
    client: Union[httpx.Client, httpx.AsyncClient],
    trace_id: str,
    upload_url: str,
    db_path: Path,
    max_retries: int = 3,
    initial_backoff: float = 1.0,
) -> Optional[str]:
    loaded_trace = load_trace_from_db(db_path, trace_id)

    msgpack = loaded_trace[0]
    compressed_data = compress_trace(msgpack)

    for attempt in range(max_retries):
        try:
            if isinstance(client, httpx.AsyncClient):
                response = await client.put(
                    upload_url,
                    content=compressed_data,
                    headers={"Content-Type": "application/gzip"},
                    timeout=30.0,
                )
            else:
                response = client.put(
                    upload_url,
                    content=compressed_data,
                    headers={"Content-Type": "application/gzip"},
                    timeout=30.0,
                )

            if response.status_code != 200:
                error_message = (
                    f"Failed to upload {trace_id}.msgpack.gz. Status code: {response.status_code}\n"
                    f"{response.text}"
                )
                click.echo(click.style(error_message, fg="red"), err=True)
                return error_message

            return None  # Success case

        except (httpx.TimeoutException, httpx.RequestError) as e:
            if attempt < max_retries - 1:
                backoff = initial_backoff * (2**attempt)
                error_message = f"Error uploading {trace_id}.msgpack.gz (attempt {attempt + 1}/{max_retries}): {str(e)}. Retrying in {backoff:.2f} seconds..."
                click.echo(click.style(error_message, fg="yellow"), err=True)
                if isinstance(client, httpx.AsyncClient):
                    await asyncio.sleep(backoff)
                else:
                    time.sleep(backoff)
            else:
                error_message = f"Failed to upload {trace_id}.msgpack.gz after {max_retries} attempts: {str(e)}"
                click.echo(click.style(error_message, fg="red"), err=True)
                return error_message

    return f"Failed to upload {trace_id}.msgpack.gz after {max_retries} attempts"


def sync_ci_upload(traces: List[Dict[str, Any]], db_path: Path):
    with httpx.Client() as client:
        with click.progressbar(traces, label="Uploading traces", show_pos=True) as bar:
            for trace in bar:
                result = asyncio.run(
                    upload_trace(
                        client=client,
                        trace_id=trace["id"],
                        upload_url=trace["upload_url"],
                        db_path=db_path,
                    )
                )
                if result:
                    click.echo(f"\n{result}", err=True)


async def async_ci_upload(traces: List[Dict[str, Any]], db_path: Path):
    async with httpx.AsyncClient(limits=httpx.Limits(max_connections=20)) as client:
        tasks = [
            upload_trace(
                client=client,
                trace_id=trace["id"],
                upload_url=trace["upload_url"],
                db_path=db_path,
            )
            for trace in traces
        ]

        bar: click._termui_impl.ProgressBar[Any]
        with click.progressbar(
            length=len(tasks), label="Uploading traces", show_pos=True
        ) as bar:
            results = []
            for task in asyncio.as_completed(tasks):
                result = await task
                bar.update(1)
                if result:
                    results.append(result)
                    click.echo(f"\n{result}", err=True)


@ci.command("upload")
@click.option("--sync", is_flag=True, help="Use synchronous upload instead of async")
def ci_upload(sync: bool):
    """
    Upload all traces in the local Kolo db to Kolo Cloud
    """
    db_path = setup_db()
    auth_token = os.environ["KOLO_TOKEN"]

    traces_in_local_db = list_traces_from_db(db_path, count=10000)

    trace_group_id = create_trace_group()

    traces_registered = []
    page_size = 500

    for i in range(0, len(traces_in_local_db), page_size):
        page = traces_in_local_db[i : i + page_size]

        base_url = os.environ.get("KOLO_BASE_URL", "https://my.kolo.app")
        response = httpx.post(
            f"{base_url}/api/traces/bulk",
            headers={"Authorization": f"Bearer {auth_token}"},
            json={
                "trace_group_id": trace_group_id,
                "traces": [{"id": trace[0], "recorded_at": trace[1]} for trace in page],
            },
        )

        if response.status_code != 201:
            error_message = f"Failed to bulk create traces. Status code: {response.status_code}\n{response.text}"
            click.echo(click.style(error_message, fg="red"), err=True)
            raise click.Abort()

        traces_registered.extend(response.json()["traces"])

        click.echo(
            f"Registered {min(i + page_size, len(traces_in_local_db))} of {len(traces_in_local_db)} traces"
        )

    if sync:
        sync_ci_upload(traces=traces_registered, db_path=db_path)
    else:
        asyncio.run(async_ci_upload(traces=traces_registered, db_path=db_path))


@cli.command()
def mcp():
    """
    Start the Kolo MCP server.
    """
    import asyncio

    from .mcp_server import mcp

    asyncio.run(mcp.run_stdio())


if __name__ == "__main__":
    cli()  # pragma: no cover



from __future__ import annotations

import datetime
import json
import mimetypes
import os
import xml.etree.ElementTree as ET
from itertools import chain
from pathlib import Path
from typing import Optional, Tuple
from urllib.request import urlopen

import platformdirs

from .. import kv
from ..config import create_kolo_directory, load_config
from ..db import (
    db_connection,
    delete_traces_before,
    delete_traces_by_id,
    load_trace_from_db,
    setup_db,
    vacuum_db,
)
from ..generate_tests import build_test_context, create_test_plan
from ..open_in_pycharm import open_in_pycharm
from ..settings import WebappSettings, get_webapp_settings, save_webapp_settings


def latest_traces(show_next, anchor=None):
    db_path = setup_db()
    with db_connection(db_path) as connection:
        needs_reversed_order = False
        reached_top = False

        if anchor is not None:
            limit = abs(show_next)

            # Positive show_next value means we're going back in time (loading _more_ traces),
            # negative value is for going forward in time (loading traces the user has previously seen before going back).

            if show_next > 0:
                # going back in time, trying to access older traces than the anchor

                query = "SELECT id FROM traces WHERE id < ? ORDER BY id desc LIMIT ?"
                cursor = connection.execute(query, (anchor, limit))
                rows = cursor.fetchall()
            else:
                # going forward in time, trying to access newer traces than the anchor

                query = "SELECT id FROM traces WHERE id > ? ORDER BY id LIMIT ?"
                needs_reversed_order = True
                # In order to get 10 newer traces, they need to be sorted in ascending order.
                # They have to be reversed later because the endpoint should always return traces from newest to oldest.

                cursor = connection.execute(query, (anchor, limit))
                rows = cursor.fetchall()

                if len(rows) < abs(limit):
                    # If there are less than 10 newer traces, we need to fetch some older traces to fill up the response.
                    cursor = connection.execute(
                        "SELECT id FROM traces ORDER BY id desc LIMIT ?", (abs(limit),)
                    )
                    rows = cursor.fetchall()

                    reached_top = True
                    needs_reversed_order = False
        else:
            # not a pagination request, we just want N latest traces

            cursor = connection.execute(
                "SELECT id FROM traces ORDER BY id desc LIMIT ?", (show_next,)
            )

            rows = cursor.fetchall()
            reached_top = True

    traces = list(chain.from_iterable(rows))

    if needs_reversed_order:
        traces = traces[::-1]

    return {"traces": traces, "isTop": reached_top}


def get_trace(trace_id):
    db_path = setup_db()
    return load_trace_from_db(db_path, trace_id)


def delete_trace(trace_id):
    db_path = setup_db()
    count = delete_traces_by_id(db_path, (trace_id,))
    return {"deleted": count}


def generate_test(trace_id):
    test_class = "MyTestCase"
    test_name = "test_my_view"

    config = load_config()
    context = build_test_context(
        trace_id, test_class=test_class, test_name=test_name, config=config
    )
    plan = create_test_plan(config, context)

    return {"test_code": plan.render(), "plan": plan.as_json()}


def get_project_root() -> Path:
    """Returns the absolute path of the project."""
    # Daniel: I'm not sure if this is correct, it just works on a simple
    # Django project.
    kolo_dir = create_kolo_directory()
    project_folder = kolo_dir.parent
    return project_folder


def read_source_file(file_path):
    assert file_path.endswith(".py")

    if not os.path.isabs(file_path):
        project_path = get_project_root()
        file_path = os.path.join(project_path, file_path)

    with open(file_path, "r") as f:
        content = f.read()

    return {"path": file_path, "content": content}


def delete_old_traces_and_vacuum():
    db_path = setup_db()
    delete_traces_before(db_path, datetime.datetime.now() - datetime.timedelta(days=30))
    vacuum_db(db_path)
    last_vacuumed = datetime.datetime.now().isoformat()
    kv.set("last_vacuumed", last_vacuumed)
    return last_vacuumed


def get_user_data_dir() -> Path:
    # Get the appropriate user data directory for the current platform
    user_data_dir = Path(platformdirs.user_data_dir(appname="kolo", appauthor="kolo"))

    # Check if the directory exists, create it if it doesn't
    if not os.path.exists(user_data_dir):
        try:
            os.makedirs(user_data_dir)
        except OSError:
            pass

    # Check if the directory is writable.
    # Important in order not to get stuck reading a directory we can never write to.
    if os.access(user_data_dir, os.W_OK):
        return user_data_dir
    # Otherwise, use the Kolo project directory.
    # User info won't be persisted across projects.
    else:
        return create_kolo_directory()


def get_user_data_filepath() -> Path:
    return get_user_data_dir() / Path("user-info.json")


def get_user_info() -> dict:
    file_path = get_user_data_filepath()
    try:
        with open(file_path, "r") as file:
            data = json.load(file)
            return data
    except OSError:
        return {}


def is_valid_email(email: str) -> bool:
    if not email:
        return False
    handle, at, domain = email.partition("@")
    return bool(handle) and bool(at) and bool(domain)


def save_user_info(data: dict):
    file_path = get_user_data_filepath()
    with open(file_path, "w") as file:
        json.dump(data, file, indent=4)


def handle_static_file(static_path: str) -> Tuple[bytes, str]:
    """Handle static file requests, returns (file_contents, mime_type)"""
    source_dir = os.path.dirname(__file__)
    static_file = f"{source_dir}/static/{static_path}"

    with open(static_file, "rb") as f:
        contents = f.read()

    mime_type, _ = mimetypes.guess_type(static_path)
    return contents, mime_type or "application/octet-stream"


def handle_open_editor(file_path: str, editor: str) -> None:
    """Handle opening files in external editors"""
    assert file_path.endswith(".py")
    assert editor == "pycharm"  # Currently only support PyCharm

    if not os.path.isabs(file_path):
        project_path = get_project_root()
        file_path = os.path.join(project_path, file_path)

    open_in_pycharm(file_path)


def save_settings_data(settings_data: WebappSettings) -> None:
    """Save webapp settings"""
    save_webapp_settings(settings_data)


def handle_user_info_update(email: str) -> None:
    """Handle user info updates with validation"""
    if not is_valid_email(email):
        raise ValueError("Invalid email")
    save_user_info({"email": email})


def get_latest_version_info() -> str:
    """Get latest version from PyPI RSS feed"""
    # We have an implementation of this in TypeScript (see extension/version.ts).
    # But we are re-implementing it here because PyPI doesn't support CORS, so we
    # can't use that implementation on the client-side.

    url = "https://pypi.org/rss/project/kolo/releases.xml"
    with urlopen(url) as response:
        rss_content = response.read()

    root = ET.fromstring(rss_content)
    channel = root.find("channel")
    if channel is None:
        raise ValueError("Could not find channel in RSS feed")

    first_item = channel.find("item")
    if first_item is None:
        raise ValueError("Could not find item in RSS feed")

    title = first_item.find("title")
    if title is None or title.text is None:
        raise ValueError("Could not extract Kolo version from RSS feed")

    return title.text


def check_vacuum_needed() -> Tuple[bool, Optional[str]]:
    """Check if vacuum is needed and perform it if necessary"""
    did_vacuum = False
    try:
        last_vacuumed = kv.get_value("last_vacuumed")
    except KeyError:
        last_vacuumed = None

    settings = get_webapp_settings()
    if settings["auto_delete_old_traces"]:
        one_day_ago = datetime.datetime.now() - datetime.timedelta(days=1)
        if (
            last_vacuumed is None
            or datetime.datetime.fromisoformat(last_vacuumed) < one_day_ago
        ):
            last_vacuumed = delete_old_traces_and_vacuum()
            did_vacuum = True

    return did_vacuum, last_vacuumed
